"""This module contains the definition of a manifest class.

A manifest object should be instantiated for each processing step of the data
preprocessing and augmentation of the Impresso project.  
"""

import copy
import json
import logging
import os
from time import strftime
from typing import Any, Union, Optional

from git import Repo

from impresso_commons.utils.s3 import get_storage_options, upload_to_s3
from impresso_commons.utils.utils import validate_against_schema
from impresso_commons.versioning.data_statistics import (
    NewspaperStatistics,
    DataStatistics,
)
from impresso_commons.versioning.helpers import (
    DataStage,
    read_manifest_from_s3,
    validate_stage,
    clone_git_repo,
    write_and_push_to_git,
    write_dump_to_fs,
    get_head_commit_url,
    increment_version,
    validate_version,
    init_media_info,
    media_list_from_mft_json,
    read_manifest_from_s3_path,
)

logger = logging.getLogger(__name__)

GIT_REPO_SSH_URL = "git@github.com:impresso/impresso-data-release.git"
REPO_BRANCH_URL = "https://github.com/impresso/impresso-data-release/tree/{branch}"

IMPRESSO_STORAGEOPT = get_storage_options()


class DataManifest:

    def __init__(
        self,
        data_stage: Union[DataStage, str],
        s3_output_bucket: str,  # including partition
        # str path to the local repo corresponding to the code used for processing
        git_repo: str | Repo,
        temp_dir: str,
        # S3 bucket of the previous stage, None if stage='canonical'
        s3_input_bucket: Optional[str] = None,
        staging: Optional[bool] = None,
        # to directly provide the next version
        new_version: Optional[str] = None,
        # to indicate if patch in later stages
        is_patch: Optional[bool] = False,
        # to indiate patch in canonical/rebuilt
        patched_fields: Union[dict[str, list[str]], list[str], None] = None,
        # directly provide the s3 path of the manifest to use as base
        previous_mft_path: Optional[str] = None,
        only_counting: Optional[bool] = False,
        notes: Optional[str] = None,
        push_to_git: Optional[bool] = False,
    ) -> None:

        # TODO when integrating radio data: init a media_type attribute and add RadioStatistics.

        self.stage = validate_stage(data_stage)  # update once all stages are final
        self.input_bucket_name = s3_input_bucket
        self.only_counting = only_counting
        self.modified_info = False
        self.push_to_git = push_to_git

        # s3_output_bucket is the path to actual data partition
        s3_output_bucket = s3_output_bucket.replace("s3://", "")
        if "/" in s3_output_bucket:
            self.output_bucket_name = s3_output_bucket.split("/")[0]
            self.output_s3_partition = "/".join(s3_output_bucket.split("/")[1:])
        else:
            # for data preparation, the manifest is at the top level of the bucket
            self.output_bucket_name, self.output_s3_partition = s3_output_bucket, None

        self.temp_dir = temp_dir
        os.makedirs(temp_dir, exist_ok=True)
        self.notes = notes

        # attributes relating to GitHub
        self.branch = self._get_output_branch(staging)
        # get code version used for processing.
        self.commit_url = get_head_commit_url(git_repo)

        # init attributes of previous manifest
        self._prev_mft_s3_path = previous_mft_path
        self.prev_version = None

        # init attribute of input_manifest
        self.input_manifest_s3_path = None

        # if user already knows the target version
        self.version = None if new_version is None else validate_version(new_version)

        # if update is a patch, patched fields should be provided
        # either as list of fields or dict mapping each media to its patched fields
        self.patched_fields = patched_fields
        self.is_patch = is_patch or (patched_fields is not None)

        # dict mapping from title to year to np_stat,
        # where the statistics will be aggregated during the processing
        self._processing_stats = {}
        self.manifest_data = {}
        self._generation_date = None

        logger.info("DataManifest for %s stage successfully initialized.", self.stage)

    @property
    def _input_stage(self) -> DataStage:
        """Get the DataStage associated to the input data of this manifest.

        Returns:
            DataStage: The input DataStage.
        """
        if self.stage in [
            DataStage.REBUILT,
            DataStage.CANONICAL,
            DataStage.PASSIM,
            DataStage.MYSQL_CIS,
        ]:
            # datastages that directly follow or use canonical data
            return DataStage.CANONICAL
        if self.stage in [DataStage.TEXT_REUSE]:
            # text reuse uses passim rebuilt text
            return DataStage.PASSIM
        # all other stages use the rebuilt as input
        return DataStage.REBUILT

    @property
    def _manifest_filename(self) -> str:
        """Get the manifest filename based on convention stage-value_vM-m-p.json.

        Returns:
            str: The manifest filename.
        """
        if self.version is None:
            logger.warning("The manifest name is only available once the version is.")
            return ""

        return f"{self.stage.value}_{self.version.replace('.', '-')}.json"

    @property
    def output_mft_s3_path(self) -> str:
        """Get this manifest's output S3 path based on its output bucket.

        The manifest will be uploaded to the S3 bucket and partition corresponding
        to the value provided for its input argument `s3_output_bucket`.
        If the versison attribute for this manifest is not defined, the S3 output
        path cannot be provided and the empty string will be returned.

        Returns:
            str: Full S3 path of this manifest if the version is already defined,
                the empty string otherwise.
        """
        if self.version is None:
            logger.warning(
                "The manifest s3 path is only available once the version is."
            )
            return ""

        if self.output_s3_partition is not None:
            s3_path = os.path.join(self.output_s3_partition, self._manifest_filename)
        else:
            s3_path = self._manifest_filename

        full_s3_path = os.path.join("s3://", self.output_bucket_name, s3_path)

        # sanity check
        if (
            self._prev_mft_s3_path is not None
            and self.output_bucket_name in self._prev_mft_s3_path
        ):
            assert (
                self._prev_mft_s3_path.split(f"/{self.stage.value}_v")[0]
                == full_s3_path.split(f"/{self.stage.value}_v")[0]
            ), "Mismatch between s3 path of previous & current version of manifest."

        return full_s3_path

    def _get_output_branch(self, for_staging: Optional[bool]) -> str:
        """Get the git repository branch on which to add the manifest once generated-

        The logic to choose the branch is the following:
            - If no `for_staging` argument was provided, only the output bucket name is used
                to infer the branch.
                - If "final" -> "master", and "staging" -> "staging"
                - If the stage is not in the bucket name, use "staging" branch by default
            - If the `for_staging` argument was provided, it can override the result.

        Hence, "master" branch is chosen only if:
            - `for_staging` was defined and False
            - `final` was in the output bucket name and `for_staging` was None

        Args:
            for_staging (Optional[bool]): Whether this manifest is meant for staging.

        Returns:
            str: Branch to use on the repo, either "staging" or "master"
        """
        staging_out_bucket = "staging" in self.output_bucket_name
        final_out_bucket = "final" in self.output_bucket_name

        if for_staging is None:
            for_staging = not (staging_out_bucket or final_out_bucket)

        return "staging" if staging_out_bucket or for_staging else "master"

    def _get_prev_version_manifest(self) -> Union[dict[str, Any], None]:
        """Find and return the previous version manifest of this DataStage and bucket.

        Will try to find the previous manifest in the provided S3 output bucket
        partition, unless `self._prev_mft_s3_path` is defined, in which case it reads
        it from S3 and returns it.

        Returns:
            Union[dict[str, Any], None]: The previous version of the manifest,
                if available, otherwise None.
        """
        # previous version manifest is in the output bucket, except when:
        # _prev_mft_s3_path is defined upon instantiation, then use it directly
        logger.debug("Reading the previous version of the manifest from S3.")
        if self._prev_mft_s3_path is None:
            (self._prev_mft_s3_path, prev_v_mft) = read_manifest_from_s3(
                self.output_bucket_name, self.stage, self.output_s3_partition
            )
        else:
            prev_v_mft = read_manifest_from_s3_path(self._prev_mft_s3_path)

        if self._prev_mft_s3_path is None or prev_v_mft is None:
            logger.info(
                "No existing previous version of this manifest. Version will be v0.0.1"
            )
            return None

        self.prev_version = prev_v_mft["mft_version"]

        return prev_v_mft

    def _get_current_version(self, addition: bool = False) -> str:
        """Get the current version of manifest for the data stage.

        Versions are of format vM.m.p (=Major.minor.patch), where:
            - M changes in the case of additions to the collection (new “title-year” keys).
            - m when modifications are made to existing data (keeping the same ‘title-year’
                keys,by re-ingestion (not field-specific modifications).
            - p when a patch or small fix is made, modifying one or more fields for several
                titles, leaving the rest of the data unchanged.
                When `self.is_patch` is True or `self.patched_fields` is defined.
                Also when `self.only_counting` is set to True and no counts changed since
                the previous manifest version.

        If no previous version exists, the version returned will be v0.0.1.

        Args:
            addition (bool, optional): Whether new data was added during the processing.
                Defaults to False.

        Returns:
            str: The current version string.
        """
        if self.prev_version is None:
            # First manifest for this data stage.
            return "v0.0.1"

        logger.debug(
            "Addition: %s, self.is_patch: %s, self.only_counting: %s, self.modified_info: %s",
            addition,
            self.is_patch,
            self.only_counting,
            self.modified_info,
        )

        if addition:
            # any new title-year pair was added during processing
            logger.info("Addition is True, major version increase.")
            return increment_version(self.prev_version, "major")

        if self.is_patch or self.patched_fields is not None:
            # processing is a patch
            logger.info("Computation is patch, patch version increase.")
            return increment_version(self.prev_version, "patch")

        if (
            self.only_counting is not None
            and self.only_counting
            and not self.modified_info
        ):
            # manifest computed to count contents of a bucket
            # (eg. after a copy from one bucket to another)
            logger.info("Only counting, and no modified info, patch version increase.")
            return increment_version(self.prev_version, "patch")

        # modifications were made by re-ingesting/re-generating the data, not patching
        logger.info(
            "No additional keys or patch, but modified info, minor version increase."
        )
        return increment_version(self.prev_version, "minor")

    def _get_input_data_overall_stats(self) -> list[dict[str, Any]]:
        """Return the `"overall_statistics"` of the input manifest or an empty list.

        The input manifest is the manifest versioning the data used as input to the
        processing used to generate data of this manifest's `DataStage`.
        If `s3_input_bucket` is defined, find and read this manifest on S3, and extract
        its `"overall_statistics"`.
        When `self.stage=DataStage.CANONICAL`, no previous stage exists, so the empty
        list will be returned.
        For other stages, the `"overall_statistics"` for all upstream stages will be
        contained in the returned list.

        Returns:
            list[dict[str, Any]]: Input manifest's `"overall_statistics"` or empty list.
        """
        # reading the input manifest only if the input s3 bucket is defined
        if self.input_bucket_name is not None:
            logger.debug("Reading the input data's manifest from S3.")

            # only the rebuilt uses the canonical as input
            (self.input_manifest_s3_path, input_v_mft) = read_manifest_from_s3(
                self.input_bucket_name.replace("s3://", ""), self._input_stage
            )

            if input_v_mft is not None:
                # assert self.input_manifest_s3_path == input_v_mft["mft_s3_path"]
                # fetch the overall statistics from the input data (it's a list!)
                if self.stage != DataStage.CANONICAL:
                    return input_v_mft["overall_statistics"]
        return []

    def _get_out_path_within_repo(
        self,
        folder_prefix: str = "data-processing-versioning",
        stage: Optional[DataStage] = None,
    ) -> str:
        """Get the output path within the "impresso-data-release" repository.


        Args:
            folder_prefix (str, optional): Folder prefix to use within the repository.
                Defaults to "data-processing-versioning".
            stage (Optional[DataStage], optional): DataStage to consider if not
                `self.stage`. Defaults to None.

        Returns:
            str: The output path within the repository based on `self.stage`.
        """
        stage = stage if stage is not None else self.stage
        if stage in ["canonical", "rebuilt", "passim", "evenized-rebuilt"]:
            sub_folder = "data-preparation"
        elif "solr" in stage or "mysql":
            sub_folder = "data-ingestion"
        else:
            sub_folder = "data-processing"

        return os.path.join(folder_prefix, sub_folder)

    def validate_and_export_manifest(
        self,
        push_to_git: bool = False,
        commit_msg: Optional[str] = None,
    ) -> bool:
        """Validate the current manifest against a schema and export it (s3 and Git).

        This function will always upload the generated manifest to S3, using a path
        constructed based on `self.output_bucket_name` and the DataStage.

        If `push_to_git` is True, by default the commit message used will be
        "Add generated manifest file {filename}." It can be overriden.

        Note:
            If a problem occurs when pushing to Git, a critical message will be logged,
            but it won't modify or alter the upload of the manifest to S3.

        Args:
            push_to_git (bool, optional): Whether to also push the generated manifest to
                GitHub (impresso/impresso-data-release). Defaults to False.
            commit_msg (Optional[str], optional): Commit message to override the
                default message. Defaults to None.

        Returns:
            bool: Whether the upload to s3 was successful.
        """
        msg = "Validating and exporting manifest to s3"

        # validate the manifest against the schema
        validate_against_schema(self.manifest_data)

        if push_to_git:
            # clone the data release repository locally if not for debug
            out_repo = clone_git_repo(self.temp_dir, branch=self.branch)

            logger.info("%s and GitHub!", msg)
        else:
            out_repo = None
            logger.info("%s!", msg)

        manifest_dump = json.dumps(self.manifest_data, indent=4)

        mft_filename = self._manifest_filename

        # if out_repo is None, in debug mode
        if push_to_git:
            # write file and push to git
            local_path_in_repo = self._get_out_path_within_repo()
            pushed, out_file_path = write_and_push_to_git(
                manifest_dump,
                out_repo,
                local_path_in_repo,
                mft_filename,
                commit_msg=commit_msg,
            )
            if not pushed:
                logger.critical(
                    "Push manifest to git manually using the file added on S3: \ns3://%s/%s.",
                    self.output_bucket_name,
                    out_file_path,
                )
        else:
            # for debug purposes, write in temp dir and not in git repo
            out_file_path = write_dump_to_fs(manifest_dump, self.temp_dir, mft_filename)

        if self.output_s3_partition is not None:
            # add the path within the bucket (partition) to the manifest file
            mft_filename = os.path.join(self.output_s3_partition, mft_filename)

        return upload_to_s3(out_file_path, mft_filename, self.output_bucket_name)

    def get_count_keys(self) -> list[str]:
        """Get the list of count keys for this manifest's media dict.

        TODO when integrating radio data: init RadioStatistics instead.

        Returns:
            list[str]: Count keys corresponding to this manifest's DataStage.
        """
        return NewspaperStatistics(self.stage, "year").count_keys

    def init_yearly_count_dict(self) -> dict[str, int]:
        """Initialize new newspaper statistics counts for this manifest.

        TODO when integrating radio data: init RadioStatistics instead.

        Returns:
            dict[str, int]: Initialized counts for this manifest.
        """
        return NewspaperStatistics(self.stage, "year").init_counts()

    def _log_failed_action(self, title: str, year: str, action: str) -> None:
        """Log and add to the notes that an action on the counts/statistics failed.

        Args:
            title (str): Media title on/with which this failure took place.
            year (str): Year on/with which this failure took place.
            action (str): Specific action during which failure took plave.
        """
        failed_note = f"{title}-{year}: {action} provided counts failed (invalid)."

        logger.warning(" ".join([failed_note, "Adding information to notes."]))
        self.append_to_notes(failed_note, to_start=False)

    def _init_yearly_stats(
        self, title: str, year: str, counts: dict[str, int]
    ) -> tuple[NewspaperStatistics, bool]:
        """Initialize the stats fro a given title and year given precomputed counts.

        TODO when integrating radio data: init RadioStatistics instead.

        Args:
            title (str): Media title to add the stats to.
            year (str): Year to which the stats correspond.
            counts (dict[str, int]): Precomputed counts used directly to initialize.

        Returns:
            tuple[NewspaperStatistics, bool]: The initiailized stats and whether this
                initialization was successful.
        """
        logger.debug("Initializing counts for %s-%s", title, year)
        elem = f"{title}-{year}"
        np_stats = NewspaperStatistics(self.stage, "year", elem, counts=counts)
        success = True
        # if the created stats don't have the initialized values
        if np_stats.counts != counts:
            success = False
            self._log_failed_action(title, year, "initializing with")

        return np_stats, success

    def has_title_year_key(self, title: str, year: str) -> bool:
        """Verify whether the provided title and year have been processed.

        Args:
            title (str): Media title to check.
            year (str): Year to check.

        Returns:
            bool: True if the title-year pair has instantiated counts, false otherwise.
        """
        if title in self._processing_stats:
            return year in self._processing_stats[title]

        return False

    def _modify_processing_stats(
        self, title: str, year: str, counts: dict[str, int], adding: bool = True
    ) -> bool:
        """Update or replace the processing stats for a given title-year pair.

        If the provided title-year key was not associated with any value in
        `processing_stats`, addition and replacement are the same: initialization.

        Args:
            title (str): Media title which should be updated with the new counts.
            year (str): Year which should be updated with the new counts.
            counts (dict[str, int]): Counts to use to add or replace current stats.
            adding (bool, optional): Whether this update should be an addition or
                replacement of any existing counts. Defaults to True.

        Returns:
            bool: True if the modification was a success, False otherwise.
        """
        # check if title/year pair is already in processing stats
        if self.has_title_year_key(title, year):
            # if title in self._processing_stats:
            #    if year in self._processing_stats[title]:
            success = self._processing_stats[title][year].add_counts(
                counts, replace=(not adding)
            )
            if not success:
                action = "adding" if adding else "replacing with"
                self._log_failed_action(title, year, action)

            # notify user of outcome
            return success

        if title not in self._processing_stats:
            logger.debug("Adding first stats for %s", title)
            self._processing_stats[title] = {}
        # initialize new statistics for this title-year pair:
        self._processing_stats[title][year], success = self._init_yearly_stats(
            title, year, counts
        )

        return success

    def add_by_ci_id(self, ci_id: str, counts: dict[str, int]) -> bool:
        """Add new counts corresponding to a specific content-item ID.

        Args:
            ci_id (str): Content-item canonical ID to which the counts correspond.
            counts (dict[str, int]): Counts corresponding to that ID.

        Returns:
            bool: True if the processing stats' update was successful, False otherwise.
        """
        title, year = ci_id.split("-")[0:2]
        return self._modify_processing_stats(title, year, counts)

    def add_by_title_year(self, title: str, year: str, counts: dict[str, int]) -> bool:
        """Add new counts corresponding to a specific media title and year.

        Args:
            title (str): Media title to which the counts correspond.
            year (str): Year to which the counts correspond.
            counts (dict[str, int]): Counts corresponding to that title and year.

        Returns:
            bool: True if the processing stats' update was successful, False otherwise.
        """
        return self._modify_processing_stats(title, str(year), counts)

    def add_count_list_by_title_year(
        self, title: str, year: str, all_counts: list[dict[str, int]]
    ) -> bool:
        """Add a list of new counts corresponding to a specific media title and year.

        Args:
            title (str): Media title to which the counts correspond.
            year (str): Year to which the counts correspond.
            all_counts (list[dict[str, int]]): Lsit of counts for that title and year.

        Returns:
            bool: True if all the updates were successful, False otherwise.
        """
        return all(
            self._modify_processing_stats(title, str(year), c) for c in all_counts
        )

    def replace_by_ci_id(self, ci_id: str, counts: dict[str, int]) -> bool:
        """Replace the current counts for a CI id's title-year pair with new ones.

        Warning:
            This operation will overwrite any current counts corresponding to the
            media title and year of the provided content-item ID. If the goal isn't
            to overwrite these counts, `add_by_ci_id` is better suited.

        Args:
            ci_id (str): Content-item canonical ID to which the counts correspond.
            counts (dict[str, int]): Counts for that ID to overwrite current counts with.

        Returns:
            bool: True if the stats' modification was successful, False otherwise.
        """
        title, year = ci_id.split("-")[0:2]
        logger.warning(
            "The counts for %s-%s will be replaced by %s (corresponding to id: %s)",
            title,
            year,
            counts,
            ci_id,
        )
        return self._modify_processing_stats(title, year, counts, adding=False)

    def replace_by_title_year(
        self, title: str, year: str, counts: dict[str, int]
    ) -> bool:
        """Replace the current counts for a given title-year pair with new ones.

        Warning:
            This operation will overwrite any current counts corresponding to the
            media title and year of the provided content-item ID. If the goal isn't
            to overwrite these counts, `add_by_title_year` is better suited.

        Args:
            title (str): Media title to which the counts correspond.
            year (str): Year to which the counts correspond.
            counts (dict[str, int]): Counts for that ID to overwrite current counts with.

        Returns:
            bool: True if the stats' modification was successful, False otherwise.
        """
        logger.warning(
            "The counts for %s-%s will be replaced by %s", title, year, counts
        )
        return self._modify_processing_stats(title, str(year), counts, adding=False)

    def append_to_notes(self, contents: str, to_start: bool = True) -> None:
        """Append a string content to the manifest notes, initialize them if needed.

        Args:
            contents (str): Text to add to the manifest notes.
            to_start (bool, optional): Whether the contents should be added to the
                start of the notes instead of the end. Defaults to True.
        """
        if self.notes is None:
            self.notes = contents
        else:
            new_notes = [contents, self.notes] if to_start else [self.notes, contents]
            print(new_notes)
            # there is no way of going to a new line, so separate with a hyphen
            self.notes = " — ".join(new_notes)

    def new_media(self, title: str) -> dict[str, Any]:
        """Add a new media dict to the media list, given its title.

        By default, this means the update information will be the following:
            - "update_type": "addition"
            - "update_level": "title"
            - "updated_years": [] # all represented years will be new
            - "updated_fields": [] # all fields will be new

        Args:
            title (str): Media title for which to add a new media.

        Returns:
            dict[str, Any]: _description_
        """
        # adding a new media means by default addition update type and title update level.
        logger.info("Creating new media dict for %s.", title)
        media = {
            "media_title": title,
            "last_modification_date": self._generation_date,
        }
        media.update(init_media_info(fields=self.patched_fields))
        media.update(
            {
                "code_git_commit": self.commit_url,
                "media_statistics": [],
                "stats_as_dict": {},
            }
        )

        return media

    def define_update_info_for_title(
        self, processed_years: set[str], prev_version_years: set[str]
    ) -> dict[str, Union[str, list]]:
        """Define a title's update info from the previous and newly updated years.

        The update information for a given title corresponds to four keys, for which
        the values provide information about what modifications took place during the
        processing this manifest is documenting.

        They are defined based on various values:
            - `self.patched_fields`: fields updated during the processing (eg. for a patch).
            - `processed_years` and `prev_version_years`

        Four cases exist:
            - All newly processed years were in the previous version
                -> full title update, only modification.
            - Part of the previous years were updated, and no newly added years:
                -> year-specific update, where all modified years will be listed.
            - All previous years were updated, and new years were added:
                -> full title update with addition.
            - Part of the previous years were updated, and new years were added:
                -> year-specific update, with addition.

        Args:
            processed_years (set[str]): Years for which statistics were computed for
                this manifest.
            prev_version_years (set[str]): Years for which statistics has already been
                computed for the previous version of this manifest.

        Returns:
            dict[str, Union[str, list]]: New update info dict for the given title.
        """
        new_info = {"last_modification_date": self._generation_date}

        if processed_years == prev_version_years:
            # exactly all previous years were updated, none added: modification, full title
            new_info.update(init_media_info(add=False, fields=self.patched_fields))

        elif processed_years & prev_version_years == processed_years:
            # part of the previous years were updated, no new years added: modification, yearly
            new_info.update(
                init_media_info(
                    add=False,
                    full_title=False,
                    years=list(map(str, processed_years)),
                    fields=self.patched_fields,
                )
            )
        elif processed_years & prev_version_years == prev_version_years:
            # all previous years were updated, and new years added: addition, full title
            new_info.update(init_media_info(fields=self.patched_fields))
        else:
            # intersection is equal to none of the two original sets.
            # part of the previous years were updated, and new years added: addition, yearly
            new_info.update(
                init_media_info(
                    full_title=False,
                    years=list(map(str, processed_years)),
                    fields=self.patched_fields,
                )
            )
        new_info["code_git_commit"] = self.commit_url

        return new_info

    def update_media_stats(
        self, title: str, yearly_stats: dict[str, dict], old_media_list: dict[str, dict]
    ) -> Union[dict, list[str]]:
        """Update a title's media statistics given the its newly computed yearly stats.

        Note that it's actually the `old_media_list`'s contents which are updated when
        necessary.

        In addition, the value of `self.only_counting` will change the behavior:
            - When False, the computation of the manifest should follow a processing, and
              all data within the `_processing_stats` (here `yearly_stats` for 1 title) will
              be considered to have been modified (or re-generated).
            - When True, the manifest is computed to verify the contents of the data, and
              the media's information will be update only if differences in statisitics are
              found between the previous and current version.

        Args:
            title (str): Media title for which to update the media list.
            yearly_stats (dict[str, dict]): New yearly statistics for the title.
            old_media_list (dict[str, dict]): Previous version manifest' media list.

        Returns:
            Union[dict, list[str]]: Previous manifest's media list potentially updated to match
                new counts, and the list of years which have been modified
        """
        # modif_media_info = False
        # list of modified years
        modif_years = []
        logger.debug("---- INSIDE update_media_stats for %s ----", title)
        for year, stats in yearly_stats.items():

            # newly added title
            if year not in old_media_list[title]["stats_as_dict"]:
                logger.info("update_media_stats - Adding new key %s-%s.", year, title)
                print(f"update_media_stats - Adding new key {year}-{title}.")
                # modif_media_info = True
                modif_years.append(year)
                logger.debug("Setting stats for %s-%s", title, year)
                old_media_list[title]["stats_as_dict"][year] = stats

            # if self.only_counting is True, only update media info if stats changed
            elif not self.only_counting or not stats.same_counts(
                old_media_list[title]["stats_as_dict"][year]
            ):
                # modif_media_info = True
                modif_years.append(year)
                if not self.only_counting:
                    logger.debug(
                        "update_media_stats - modified stats -> Setting stats for %s-%s",
                        title,
                        year,
                    )
                old_media_list[title]["stats_as_dict"][year] = stats

        logger.info(
            "---- FINISHED update_media_stats for %s – had modifications:%s ----",
            title,
            len(modif_years) != 0,
        )
        return old_media_list, modif_years

    def generate_media_dict(self, old_media_list: dict[str, dict]) -> tuple[dict, bool]:
        """Given the previous manifest's and current statistics, generate new media dict.

        The previous version media list is updated with current processing media list:
            - Setting new modification date & git url for each modified title.
            - Compute update level & targets if not the processing is not a patch.

        From this update, also conclude on whether new data was added, informing the
        how the version should be increased: if new title-year keys exist, the "addition"
        flag will conduct to a major verison increase.

        Args:
            old_media_list (dict[str, dict]): _description_

        Returns:
            tuple[dict, bool]: _description_
        """
        addition = False
        for title, yearly_stats in self._processing_stats.items():

            # if title not yet present in media list, initialize new media dict
            if title not in old_media_list:
                # new title added to the list: addition, full title
                old_media_list[title] = self.new_media(title)
                addition = True

                # set the statistics
                old_media_list, _ = self.update_media_stats(
                    title, yearly_stats, old_media_list
                )
            else:
                prev_version_years = set(old_media_list[title]["stats_as_dict"].keys())

                # update the statistics and identify if the media info needs to change
                old_media_list, modif_years = self.update_media_stats(
                    title, yearly_stats, old_media_list
                )

                if len(modif_years) != 0:

                    # if only counting, only consider the years which were actually modified/added
                    processed_years = (
                        set(modif_years)
                        if self.only_counting
                        else set(yearly_stats.keys())
                    )

                    media_update_info = self.define_update_info_for_title(
                        processed_years,
                        prev_version_years,
                    )

                    logger.debug(
                        "define_update_info_for_title for %s, new update info: %s",
                        title,
                        media_update_info,
                    )

                    old_media_list[title].update(media_update_info)
                    print("Updated media information for %s", title)
                    logger.info("Updated media information for %s", title)
                    # keep track that media info was updated for version increase
                    self.modified_info = True

                    # if title was already present, update the information with current processing
                    if not addition and media_update_info["update_type"] == "addition":
                        # only one addition is enough
                        addition = True

        return old_media_list, addition

    def aggregate_stats_for_title(
        self, title: str, media_dict: dict[str, Any]
    ) -> tuple[dict[str, Any], NewspaperStatistics]:
        """Aggregate all stats of given title and export them to a "pretty print" dict.

        TODO once the radio data is handled, add RadioStatistics

        The `DataStatistics` objects don't display in the dict format by default,
        but need to be converted to dicts to show as desired on the final manifest.

        Args:
            title (str): Media title for which to aggregate the yearly stats.
            media_dict (dict[str, Any]): Title's media dict with formatted statistics.

        Returns:
            tuple[dict[str, Any], NewspaperStatistics]: Updated media dict and
                corresponding title-level DataStatistics object.
        """
        logger.debug("Aggregating title-level stats for %s.", title)
        title_cumm_stats = NewspaperStatistics(self.stage, "title", title)
        pretty_counts = []

        for _, np_year_stat in media_dict["stats_as_dict"].items():
            if isinstance(np_year_stat, DataStatistics):
                # newly added titles will be DataStatistics objects -> needs pretty print
                title_cumm_stats.add_counts(np_year_stat.counts)
                # include the modification date at the year level
                pretty_counts.append(
                    np_year_stat.pretty_print(modif_date=self._generation_date)
                )
            else:
                # non-modified stats will be in pretty-print dict format -> can be added directly
                title_cumm_stats.add_counts(np_year_stat["nps_stats"])
                pretty_counts.append(np_year_stat)

        # insert the title-level statistics at the "top" of the statistics
        pretty_counts.insert(0, title_cumm_stats.pretty_print())
        media_dict["media_statistics"] = pretty_counts

        return media_dict, title_cumm_stats

    def title_level_stats(
        self, media_list: dict[str, dict]
    ) -> tuple[list[DataStatistics], dict[str, dict]]:
        """Compute the title-level statistics from the new media list.

        Also removes the `stats_as_dict` field from the media list, and returns
        the media list with each NewspaperStatistics object "pretty printed".

        Args:
            media_list (dict[str, dict]): Updated media list for this manifest.

        Returns:
            tuple[list[DataStatistics], dict[str, dict]]: New title-level stats and
                media list.
        """
        full_title_stats = []
        for title, media_as_dict in media_list.items():
            # update the canonical_media_list with the new media_dict
            media_as_dict, title_cumm_stats = self.aggregate_stats_for_title(
                title, media_as_dict
            )
            # remove the stats in dict format
            del media_as_dict["stats_as_dict"]
            # save the title level statistics for the overall statistics
            full_title_stats.append(title_cumm_stats)

        return full_title_stats, media_list

    def overall_stats(self, title_stats: list[DataStatistics]) -> list[dict]:
        """Generate the overall stats and append the ones from the input manifest.

        Args:
            title_stats (list[DataStatistics]): List of all title-level statistics
                used to compute the overall stats.

        Returns:
            list[dict]: This manifest's overall stats with the ones of previous stages.
        """
        corpus_stats = NewspaperStatistics(self.stage, "corpus", "")
        for np_stats in title_stats:
            corpus_stats.add_counts(np_stats.counts)
        # add the number of titles present in corpus
        corpus_stats.add_counts({"titles": len(title_stats)})

        # add these overall counts to the ones of previous stages
        overall_stats = self._get_input_data_overall_stats()
        overall_stats.append(corpus_stats.pretty_print())

        return overall_stats

    def compute(
        self, export_to_git_and_s3: bool = True, commit_msg: Optional[str] = None
    ) -> None:
        """Perform all necessary logic to compute and construct the resulting manifest.

        This lazy behavior ensures all necessary information is ready and accessible
        when generating the manifest (in particular the `_processing_stats`).

        The steps of this computation are the following:
            - Ensure `_processing_stats` is not empty so the manifest can be computed and
              crystallize the time this function is called as the `_generation_date`.
            - Fetch the previous version of this manifest from S3, extract its media list.
            - Generate the new media list given the previous one and `_processing_stats`.
            - Compute the new title and corpus level statistics using the new media list.
            - Compute the new version based on the performed updates.
            - Define the `manifest_data` attribute corresponding to the final manifest.
            - Optionally, dump it to JSON, export it to S3 and Git.

        Args:
            export_to_git_and_s3 (bool, optional): Whether to export the final
                `manifest_data` as JSON to S3 and GitHub. Defaults to True. If False,
                `validate_and_export_manifest` can be called separately to do it.
            commit_msg (Optional[str], optional): Commit message to use instead of
                the default from `validate_and_export_manifest`. Defaults to None.
        """
        if not self._processing_stats:
            # `self._processing_stats` is empty the manifest can't be computed, directly stop.
            msg = "The manifest can't be computed without having provided statistics!"
            logger.warning(msg)

        else:
            logger.info("Starting to compute the manifest...")

            self._generation_date = strftime("%Y-%m-%d %H:%M:%S")

            #### IMPLEMENT LOGIC AND FILL MANIFEST DATA

            logger.info("Loading the previous version of this manifest if it exists.")
            # load previous version of this manifest
            prev_version_mft = self._get_prev_version_manifest()

            if prev_version_mft is not None:
                logger.info("Found previous version manifest, loading its contents.")
                # ensure a non-modified version remains
                old_mft = copy.deepcopy(prev_version_mft)
                old_media_list = media_list_from_mft_json(old_mft)

            else:
                logger.info(
                    "No previous version found, reinitializaing the media list."
                )
                # if no previous version media list, generate media list from scratch
                old_media_list = {}

                if self.only_counting:
                    logger.info(
                        "When no previous version exists, the option to only "
                        "count (only_counting=True) is not available."
                    )
                    self.only_counting = False

            logger.info("Updating the media statistics with the new information...")
            # compare current stats to previous version stats
            updated_media, addition = self.generate_media_dict(old_media_list)

            logger.info("Computing the title-level statistics...")
            full_title_stats, updated_media = self.title_level_stats(updated_media)

            logger.info("Computing the overall statistics...")
            overall_stats = self.overall_stats(full_title_stats)

            # compute current version
            self.version = self._get_current_version(addition)

            # the canonical has no input stage
            if (
                self.stage != DataStage.CANONICAL
                and self.input_manifest_s3_path is not None
            ):
                input_mft_git_path = os.path.join(
                    self._get_out_path_within_repo(stage=self._input_stage),
                    self.input_manifest_s3_path.split("/")[-1],
                )
            else:
                input_mft_git_path = None

            if self.notes is None:
                self.notes = ""

            # populate the dict with all gathered information
            self.manifest_data = {
                "mft_version": self.version,
                "mft_generation_date": self._generation_date,
                "mft_s3_path": self.output_mft_s3_path,
                "input_mft_s3_path": self.input_manifest_s3_path,
                "input_mft_git_path": input_mft_git_path,
                "code_git_commit": self.commit_url,
                "media_list": list(updated_media.values()),
                "overall_statistics": overall_stats,
                "notes": self.notes,
            }

            logger.info("%s Manifest successfully generated! %s", "-" * 15, "-" * 15)

            if export_to_git_and_s3:

                # if push_to_git is not defined and exporting directly,
                # will both upload to s3 and push to git.
                push = self.push_to_git if self.push_to_git is not None else True
                if not push:
                    logger.info(
                        "Argument export_to_git_and_s3 was set to True but push_to_git"
                        " was set to False. Exporting to S3 but not pushing to git."
                    )
                success = self.validate_and_export_manifest(push, commit_msg)

                if success:
                    logger.info(
                        "%s Manifest successfully uploaded to S3 and GitHub! %s",
                        "-" * 15,
                        "-" * 15,
                    )



#!/usr/bin/env python
# coding: utf-8
# created on 2018.03.27 using PyCharm
# project impresso-image-acquisition

import logging
import sys
import time
import datetime
from datetime import timedelta

import dask
from dask import compute, delayed
from dask.diagnostics import ProgressBar
from dask.multiprocessing import get as mp_get
import multiprocessing

logger = logging.getLogger(__name__)


def executetask(tasks, parallel_execution):
    """ Effectively run execution of tasks

    :param tasks                : list of tasks to be executed
    :type tasks                 : xx
    :param parallel_execution   :
    :type parallel_execution    : boolean ?
    """
    with ProgressBar():
        if parallel_execution:
            result = compute(*tasks, get=mp_get)
        else:
            result = compute(*tasks, get=dask.get)
    return result


def init_logger(logger, log_level, log_file):
    """Initialise the logger."""
    logger.setLevel(log_level)

    formatter = logging.Formatter(
        '%(asctime)s %(name)-12s %(levelname)-8s %(message)s'
    )

    if log_file is not None:
        fh = logging.FileHandler(filename=log_file, mode='w')
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    ch = logging.StreamHandler(sys.stdout)
    ch.setFormatter(formatter)

    logger.addHandler(ch)
    logger.info("Logger successfully initialised")

    return logger


def user_confirmation(question, default=None):
    """
    Ask a yes/no question via raw_input() and return their answer.

    @param question: a string that is presented to the user.
    @param default: the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).
    @return: True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' (or 'y' or 'n').\n")



def user_question(variable_to_confirm):
    answer = user_confirmation(f"Is [{variable_to_confirm}] the correct one to work with?", None)
    if not answer:
        logger.info(f"Variable {variable_to_confirm} not confirmed, exiting.")
        sys.exit()
    else:
        logger.info(f"Variable {variable_to_confirm} confirmed.")


def timestamp():
    """Returns an iso-formatted timestamp.

    :return: a timestamp
    :rtype: str
    """
    utcnow = datetime.datetime.utcnow()
    ts = int(utcnow.timestamp())
    d = datetime.datetime.fromtimestamp(ts)
    return d.isoformat() + "Z"


class Timer:
    """ Basic timer"""
    def __init__(self):
        self.start = time.time()
        self.intermediate = time.time()

    def tick(self):
        elapsed_time = time.time() - self.intermediate
        self.intermediate = time.time()
        return str(timedelta(seconds=elapsed_time))

    def stop(self):
        elapsed_time = time.time() - self.start
        return str(timedelta(seconds=elapsed_time))


def _get_cores():
    nb = multiprocessing.cpu_count()
    if nb >= 48:
        return nb - 10
    else:
        return nb


def chunk(list, chunksize):
    """Yield successive n-sized chunks from list."""
    for i in range(0, len(list), chunksize):
        yield list[i:i + chunksize]



#!/usr/bin/env python36
# coding: utf-8 

"""
A class to load configuration files in json format, handling different task setting.
"""
__author__ = "maudehrmann"

import logging

from impresso_commons.utils.utils import parse_json

logger = logging.getLogger(__name__)


class Base:
    """ Base class for initial loading and default checking methods"""
    def to_dict(self):
        return self.__dict__

    @classmethod
    def from_json(cls, json_file):
        logger.info(f"Loading config file {json_file}")
        config_dict = parse_json(json_file)
        return cls(config_dict)

    def check_params(self, config_dict, keys):
        if all(k in config_dict for k in keys):
            return True
        else:
            raise Exception(f"{self.__class__.__name__} misses a parameter among {keys}")

    def check_bucket(self, string, attribute):
        assert string in attribute, f"Bucket name should contain '{string}'"


class PartitionerConfig(Base):
    def __init__(self, config_dict):
        k = ['s3_bucket_rebuilt', "newspapers", 'output_dir', 'local_fs', 'keep_full']
        self.check_params(config_dict, k)

        self.bucket_name = config_dict["s3_bucket_rebuilt"]
        self.newspapers = config_dict["newspapers"]
        self.output_dir = config_dict["output_dir"]
        self.local_fs = config_dict["local_fs"]
        self.keep_full = config_dict["keep_full"]
        self.check_bucket("rebuilt", self.bucket_name)


def main():
    file = "../config/solr_ci_builder_config.example.json"
    config = PartitionerConfig.from_json(file)
    print(config.newspapers)


if __name__ == '__main__':
    main()



"""Helper functions to read, generate and write data versioning manifests.
"""

import copy
import json
import logging
import os
import re
from ast import literal_eval
from collections import Counter
from time import strptime
from typing import Any, Union, Self, Optional

# python 3.10+
from enum import StrEnum
from tqdm import tqdm

import git
from dask import dataframe as dd
import dask.bag as db
from dask.distributed import progress, Client

from impresso_commons.utils.utils import bytes_to
from impresso_commons.utils.s3 import (
    fixed_s3fs_glob,
    alternative_read_text,
    get_storage_options,
    get_boto3_bucket,
    get_s3_object_size,
)

logger = logging.getLogger(__name__)


IMPRESSO_STORAGEOPT = get_storage_options()
POSSIBLE_GRANULARITIES = ["corpus", "title", "year"]
VERSION_INCREMENTS = ["major", "minor", "patch"]


class DataStage(StrEnum):
    """Enum all stages requiring a versioning manifest.

    Each member corresponds to a data stage and the associated string is used to name
    each generated manifest accordingly.

    TODO: finalize the exact list of names and strings based on needs.
    TODO: add options for data indexing in Solr
    """

    CANONICAL = "canonical"
    REBUILT = "rebuilt"
    EVENIZED = "evenized-rebuilt"
    PASSIM = "passim"
    EMBEDDINGS = "embeddings"
    ENTITIES = "entities"
    LANGIDENT = "langident"
    LINGUISTIC_PROCESSING = "lingproc"
    OCRQA = "ocrqa"
    TEXT_REUSE = "text-reuse"
    TOPICS = "topics"
    SOLR_TEXT = "solr-ingestion-text"
    SOLR_ENTITIES = "solr-ingestion-entities"
    SOLR_EMBS = "solr-ingestion-emb"
    MYSQL_CIS = "mysql-ingestion"

    @classmethod
    def has_value(cls: Self, value: str) -> bool:
        """Check if enum contains given value

        Args:
            cls (Self): This DataStage class
            value (str): Value to check

        Returns:
            bool: True if the value provided is in this enum's values, False otherwise.
        """
        return value in cls._value2member_map_


##################################
###### VALIDATION FUNCTIONS ######


def validate_stage(
    data_stage: str, return_value_str: bool = False
) -> Union[DataStage, str, None]:
    """Validate the provided data stage if it's in the DataStage Enum (key or value).

    Args:
        data_stage (str): Data stage key or value to validate.
        return_value_str (bool, optional): Whether to return the data stage's value if
            it was valid. Defaults to False.

    Raises:
        e: The provided str is neither a data stage key nor value.

    Returns:
        DataStage | str | None: The corresponding DataStage or value string if valid.
    """
    try:
        if DataStage.has_value(data_stage):
            stage = DataStage(data_stage)
        else:
            stage = DataStage[data_stage]
        return stage.value if return_value_str else stage
    except ValueError as e:
        err_msg = f"{e} \nProvided data stage '{data_stage}' is not a valid data stage."
        logger.critical(err_msg)
        raise e


def validate_granularity(value: str) -> Optional[str]:
    """Validate that the granularity value provided is valid.

    Statistics are computed on three granularity levels:
    corpus, title and year.

    Args:
        value (str): Granularity value to validate

    Raises:
        ValueError: The provided granularity isn't one of corpus, title and year.

    Returns:
        Optional[str]: The provided value, in lower case, or None if not valid.
    """
    lower = value.lower()
    if lower in POSSIBLE_GRANULARITIES:
        return lower
    # only incorrect granularity values will not be returned
    logger.critical("Provided granularity '%s' isn't valid.", lower)
    raise ValueError


###############################
###### VERSION FUNCTIONS ######


def validate_version(v: str, regex: str = "^v([0-9]+[.]){2}[0-9]+$") -> Optional[str]:
    """Validate the provided string version against a regex.

    The provided version should be in format "vM.m.p", where M, m and p are
    integers representing respectively the Major, minor and patch version.

    Args:
        v (str): version in string format to validate.
        regex (str, optional): Regex against which to match the version.
            Defaults to "^v([0-9]+[.]){2}[0-9]+$".

    Returns:
        Optional[str]: The provided version if it's valid, None otherwise.
    """
    # accept versions with hyphens in case of mistake
    v = v.replace("-", ".")

    if re.match(regex, v) is not None:
        return v

    msg = f"Non conforming version {v} provided: ({regex}), version will be inferred."
    logger.critical(msg)
    return None


def version_as_list(version: str) -> list[int]:
    """Return the provided string version as a list of three ints.

    Args:
        version (str): String version to return as list

    Returns:
        list[int]: list of len 3 where indices respecively correspond to the
            Major, minor and patch versions.
    """
    if version[0] == "v":
        version = validate_version(version)
        start = 1
    else:
        start = 0
    sep = "." if "." in version else "-"
    return version[start:].split(sep)


def extract_version(name_or_path: str, as_int: bool = False) -> Union[str, int]:
    """Extract the version from a string filename or path.

    This function is in particular mean to extract the version from paths or filenames
    of manifests: structured as [data-stage]_vM-m-p.json.

    Args:
        name_or_path (str): Filename or path from which to extract the version.
        as_int (bool, optional): Whether to return the extracted version as int or str.
            Defaults to False.

    Returns:
        Union[str, int]: Extracted version, as int or str based on `as_int`.
    """
    # in the case it's a path
    basename = os.path.basename(name_or_path)
    version = basename.replace(".json", "").split("_")[-1]

    return int(version[1:].replace("-", "")) if as_int else version.replace("-", ".")


def increment_version(prev_version: str, increment: str) -> str:
    """Update  given version accoding to the given increment.

    When the increment is major or minor, all following numbers are reset to 0.

    Args:
        prev_version (str): Version to increment
        increment (str): Increment, can be one of major, minor and patch.

    Raises:
        e: Increment value provided is not valid.

    Returns:
        str: Vesion incremented accordingly.
    """
    try:
        incr_val = VERSION_INCREMENTS.index(increment)
        list_v = version_as_list(prev_version)
        # increase the value of the correct "sub-version" and reset the ones right of it
        list_v[incr_val] = str(int(list_v[incr_val]) + 1)
        if incr_val < 2:
            list_v[incr_val + 1 :] = ["0"] * (2 - incr_val)
        return "v" + ".".join(list_v)
    except ValueError as e:
        logger.error(
            "Provided invalid increment %s: not in %s", increment, VERSION_INCREMENTS
        )
        raise e


#####################################
###### S3 READ/WRITE FUNCTIONS ######


def find_s3_data_manifest_path(
    bucket_name: str, data_stage: str, partition: Optional[str] = None
) -> Optional[str]:
    """Find and return the latest data manifest in a given S3 bucket.

    On S3, different Data stages will be stored in different ways.
    In particular, data stages corresponding to enrichments are all placed in the
    same bucket but in different partitions.
    Data stages "canonical", "rebuilt", "evenized-rebuilt" & ones related to Solr
    are the ones where each stage has its own bucket.

    Args:
        bucket_name (str): Name of the bucket in which to look.
        data_stage (str): Data stage corresponding to the manifest to fetch.
        partition (Optional[str], optional): Partition within the bucket to look
            into. Defaults to None.

    Returns:
        Optional[str]: S3 path of the latest manifest in the bucket, None if no
            manifests were found inside.
    """
    # fetch the data stage as the naming value
    if isinstance(data_stage, DataStage):
        stage_value = data_stage.value
    else:
        stage_value = validate_stage(data_stage, return_value_str=True)

    # manifests have a json extension and are named after the format (value)
    path_filter = f"{stage_value}_v*.json"

    if partition is None and stage_value in [
        DataStage.CANONICAL.value,  # "canonical"
        DataStage.REBUILT.value,  # "rebuilt"
        DataStage.EVENIZED.value,  # "evenized-rebuilt"
        DataStage.PASSIM.value,  # "passim"
        DataStage.SOLR_TEXT.value,  # "solr-ingestion-text"
    ]:
        # manifest in top-level partition of bucket
        bucket = get_boto3_bucket(bucket_name)
        matches = fixed_s3fs_glob(path_filter, boto3_bucket=bucket)
    else:
        assert partition is not None, "partition should be provided for processed data"
        # processed data are all in the same bucket,
        # manifest should be directly fetched from path
        full_s3_path = os.path.join(bucket_name, partition, path_filter)
        print(full_s3_path)
        matches = fixed_s3fs_glob(full_s3_path)

    # matches will always be a list
    if len(matches) == 1:
        return matches[0]
    if len(matches) == 0:
        # no matches means it's the first manifest for the stage or bucket
        return None

    # if multiple versions exist, return the latest one
    return sorted(matches, key=lambda x: extract_version(x, as_int=True))[-1]


def read_manifest_from_s3(
    bucket_name: str,
    data_stage: Union[DataStage, str],
    partition: Optional[str] = None,
) -> Optional[tuple[str, dict[str, Any]]]:
    """Read and load manifest given an S3 bucket.

    Args:
        bucket_name (str): NAme of the s3 bucket to look into
        data_stage (Union[DataStage, str]): Data stage corresponding to the
            manifest to fetch.
        partition (Optional[str], optional): Partition within the bucket to look
            into. Defaults to None.

    Returns:
        tuple[str, dict[str, Any]] | tuple[None, None]: S3 path of the manifest
            and corresponding contents, if a manifest was found, None otherwise.
    """
    manifest_s3_path = find_s3_data_manifest_path(bucket_name, data_stage, partition)
    if manifest_s3_path is None:
        logger.info("No %s manifest found in bucket %s", data_stage, bucket_name)
        return None, None

    raw_text = alternative_read_text(
        manifest_s3_path, IMPRESSO_STORAGEOPT, line_by_line=False
    )

    return manifest_s3_path, json.loads(raw_text)


def read_manifest_from_s3_path(manifest_s3_path: str) -> Optional[dict[str, Any]]:
    """read and extract the contents of an arbitrary manifest,

    Args:
        manifest_s3_path (str): S3 path of the manifest to read.

    Returns:
        Optional[dict[str, Any]]: Contents of manifest if found on S3, None otherwise.
    """
    try:
        raw_text = alternative_read_text(
            manifest_s3_path, IMPRESSO_STORAGEOPT, line_by_line=False
        )
        return json.loads(raw_text)
    except FileNotFoundError as e:
        logger.error("No manifest found at s3 path %s. %s", manifest_s3_path, e)
        return None


###########################
###### GIT FUNCTIONS ######


def write_dump_to_fs(file_contents: str, abs_path: str, filename: str) -> Optional[str]:
    """Write a provided string dump to the local filesystem given its path and filename.

    TODO: Potentially moving this method to `utils.py`.

    Args:
        file_contents (str): Dumped contents in str format, ready to be written.
        abs_path (str): Local path to the directory in which the file will be.
        filename (str): Filename of the file to write, including its extension.

    Returns:
        Optional[str]: Full path of writen file, or None if an IOError occurred.
    """
    full_file_path = os.path.join(abs_path, filename)

    # write file to absolute path provided
    try:
        with open(full_file_path, "w", encoding="utf-8") as outfile:
            outfile.write(file_contents)

        return full_file_path
    except IOError as e:
        logger.error(
            "Writing dump %s to local fs triggered the error: %s", full_file_path, e
        )
        return None


def is_git_repo(path: str) -> bool:
    """Check if a directory contains a Git repository.

    Args:
        path (str): The path to the directory to be checked.

    Returns:
        bool: True if the directory contains a Git repository, False otherwise.
    """
    try:
        _ = git.Repo(path).git_dir
        return True
    except git.exc.InvalidGitRepositoryError:
        return False
    except git.NoSuchPathError:
        return False


def clone_git_repo(
    path: str, repo_name: str = "impresso/impresso-data-release", branch: str = "master"
) -> git.Repo:
    """Clone a git repository into a given path in the local file-system.

    Args:
        path (str): Path (ideally absolute) to the dir in which to clone the git repo.
        repo_name (str, optional): Full name of the git repository to clone, as it
            appears in its URL. Defaults to "impresso/impresso-data-release".
        branch (str, optional): Specific branch to clone. Defaults to "master".

    Raises:
        e: Cloning the repo failed, both using SSH and HTTPS.

    Returns:
        git.Repo: Object representing the cloned repository if it was cloned.
    """
    # WARNING: path should be absolute path!!!
    repo_ssh_url = f"git@github.com:{repo_name}.git"
    repo_https_url = f"https://github.com/{repo_name}.git"

    repo_path = os.path.join(path, repo_name.split("/")[1])

    # if the repository was already cloned, pull and return it.
    if os.path.exists(repo_path) and is_git_repo(repo_path):
        logger.info(
            "Git repository %s had already been cloned, pulling from branch %s.",
            repo_name,
            branch,
        )
        print(
            "Git repository %s had already been cloned, pulling from branch %s.",
            repo_name,
            branch,
        )
        repo = git.Repo(repo_path)
        # check if the current branch is the correct one & pull latest version
        if branch not in repo.active_branch.name:
            logger.info(
                "Switching branch from %s to %s", repo.active_branch.name, branch
            )
            print("Switching branch from %s to %s", repo.active_branch.name, branch)
            repo.git.checkout(branch)
        repo.remotes.origin.pull()

        return repo

    # try to clone using ssh, if it fails, retry with https.
    try:
        logger.info("Cloning the %s git repository with ssh.", repo_name)
        return git.Repo.clone_from(repo_ssh_url, repo_path, branch=branch)

    except git.exc.GitCommandError as e:
        err_msg = (
            f"Error while cloning the git repository {repo_name} using ssh, trying "
            f"with https. \n{e}"
        )
        logger.warning(err_msg)
    # Fallback to https
    try:
        logger.info("Cloning the %s git repository with https.", repo_name)
        return git.Repo.clone_from(repo_https_url, repo_path, branch=branch)

    except Exception as e:
        err_msg = (
            f"Error while cloning the git repository {repo_name}, it was not possible "
            f"to clone it with ssh or https. \n{e}"
        )
        logger.critical(err_msg)
        raise e


def write_and_push_to_git(
    file_contents: str,
    git_repo: git.Repo,
    path_in_repo: str,
    filename: str,
    commit_msg: Optional[str] = None,
) -> tuple[bool, str]:
    """Given a serialized dump, write it in local git repo, commit and push.

    Args:
        file_contents (str): Serialized dump of a JSON file.
        git_repo (git.Repo): Object representing the git repository to push to.
        path_in_repo (str): Relative path where to write the file.
        filename (str): Desired name for the file, including extension.
        commit_msg (Optional[str], optional): Commit message. If not defined, a
            basic message on the added manifest will be used.Defaults to None.

    Returns:
        tuple[bool, str]: Whether the process was successful and corresponding filepath.
    """
    # given the serialized dump or a json file, write it in local git repo
    # folder and push it to the given subpath on git
    local_repo_base_dir = git_repo.working_tree_dir

    git_path = os.path.join(local_repo_base_dir, path_in_repo)
    # write file in git repo cloned locally
    full_git_path = write_dump_to_fs(file_contents, git_path, filename)

    if full_git_path is not None:
        return git_commit_push(full_git_path, git_repo, commit_msg), full_git_path
    else:
        return False, os.path.join(git_path, filename)


def git_commit_push(
    full_git_filepath: str, git_repo: git.Repo, commit_msg: Optional[str] = None
) -> bool:
    """Commit and push the addition of a given file within the repository.

    TODO: make more general for non-manifest related uses?

    Args:
        full_git_filepath (str): Path to the file added to the git repository.
        git_repo (git.Repo): git.Repo object of the repository to commit and push to.
        commit_msg (Optional[str], optional): Message to use when commiting. If not
            defined, a basic message on the added manifest will be used. Defaults to None.

    Returns:
        bool: Whether the commit and push operations were successful.
    """
    # add, commit and push the file at the given path.
    filename = os.path.basename(full_git_filepath)
    # git add file
    git_repo.index.add([full_git_filepath])
    try:
        # git commit and push
        if commit_msg is None:
            commit_msg = f"Add generated manifest file {filename}."
        git_repo.index.commit(commit_msg)
        origin = git_repo.remote(name="origin")

        push_msg = f"Pushing {filename} with commit message '{commit_msg}'"
        logger.info(push_msg)
        origin.push()

        return True
    except git.exc.GitError as e:
        err_msg = f"Error while pushing {filename} to its remote repository. \n{e}"
        logger.error(err_msg)
        return False


def get_head_commit_url(repo: str | git.Repo) -> str:
    """Get the URL of the last commit on a given Git repository.

    TODO: test the function when repo is https url of repository.
    TODO: provide branch argument.

    `repo` can be one of three things:
        - a git.Repo instantiated object (if alreaday instantiated outside).
        - the local path to the git repository (previously cloned).
        - the HTTPS URL to the Git repository.

    Note:
        The returned commit URL corresponds to the one on the repository's active
        branch (master for the URL).

    Args:
        repo (str | git.Repo): local path, git.Repo object or URL of the repository.

    Returns:
        str: The HTTPS URL of the last commit on the git repository's master branch.
    """
    not_repo_url = True
    if isinstance(repo, str):
        if "https" in repo:
            # if it's the https link to the repo, get the hash of last commit
            not_repo_url = False
            commit_hash = git.cmd.Git().ls_remote(repo, heads=True).split()[0]
            raw_url = repo
        else:
            # get the actual repo if it's only the path.
            repo = git.Repo(repo)

    if not_repo_url:
        # final url of shape 'https://github.com/[orga_name]/[repo_name]/commit/[hash]'
        # warning --> commit on the repo's current branch!
        commit_hash = str(repo.head.commit)
        # url of shape 'git@github.com:[orga_name]/[repo_name].git'
        # or of shape 'https://github.com/[orga_name]/[repo_name].git'
        raw_url = repo.remotes.origin.url

    if raw_url.startswith("https://"):
        if ".git" in raw_url:
            url_end = "/".join(["", "commit", commit_hash])
            return raw_url.replace(".git", url_end)
        return "/".join([raw_url, "commit", commit_hash])

    # now list with contents ['git', 'github', 'com', [orga_name]/[repo_name], 'git']
    url_pieces = re.split(r"[@.:]", raw_url)
    # replace the start and end of the list
    url_pieces[0] = "https:/"
    url_pieces[-1] = "commit"
    # merge back the domain name and remove excedentary element
    url_pieces[1] = ".".join(url_pieces[1:3])
    del url_pieces[2]
    # add the commit hash at the end
    url_pieces.append(commit_hash)

    return "/".join(url_pieces)


##########################################
###### MEDIA LIST & STATS FUNCTIONS ######


def media_list_from_mft_json(json_mft: dict[str, Any]) -> dict[str, dict]:
    """Extract the `media_list` from a manifest as a dict where each title is a key.

    For each title, all fields from the original media list will still be present
    along with an additional `stats_as_dict` field containing a dict mapping each
    year to its specific statistics.

    As a result:
        - All represented titles are within the keys of the returned media list.
        - For each title, represented years are in the keys of its `stats_as_dict` field.

    Args:
        json_mft (dict[str, Any]): Dict following the JSON schema of a manifest from
            which to extract the media list.

    Returns:
        dict[str, dict]: Media list of given manifest, with `stats_as_dict` field.
    """
    # prevent modification or original manifest, to recover later
    manifest = copy.deepcopy(json_mft)
    new_media_list = {}
    for media in manifest["media_list"]:
        if media["media_title"] not in ["0002088", "0002244"]:
            yearly_media_stats = {
                year_stats["element"].split("-")[1]: year_stats
                for year_stats in media["media_statistics"]
                if year_stats["granularity"] == "year"
            }

            new_media_list[media["media_title"]] = media
            new_media_list[media["media_title"]]["stats_as_dict"] = yearly_media_stats
        else:
            logger.info(
                "Skipping %s as it's BL and only a sample.", media["media_title"]
            )

    return new_media_list


def init_media_info(
    add: bool = True,
    full_title: bool = True,
    years: Optional[list[str]] = None,
    fields: Optional[list[str]] = None,
) -> dict[str, Any]:
    """Initialize the media update dict for a title given relevant information.

    All the update informations are relating to the newly processed data, in
    comparison with the one computed during the last processing.

    Args:
        add (bool, optional): Whether new data was added. Defaults to True.
        full_title (bool, optional): Whether all the title's years were modified.
            Defaults to True.
        years (Optional[list[str]], optional): When `full_title`, the specific years
            which were modified/updated. Defaults to None.
        fields (Optional[list[str]], optional): List of specific fields that were
            modified/updated. Defaults to None.

    Returns:
        dict[str, Any]: Instantiated dict with the update information for a given media.
    """
    return {
        "update_type": "addition" if add else "modification",
        "update_level": "title" if full_title else "year",
        "updated_years": sorted(years) if years is not None else [],
        "updated_fields": fields if fields is not None else [],
    }


def counts_for_canonical_issue(
    issue: dict[str, Any], include_np_yr: bool = False
) -> dict[str, int]:
    """Given the canonical representation of an issue, get its counts.

    Args:
        issue (dict[str, Any]): Canonical JSON representation of an issue.
        include_np_yr (bool, optional): Whether the newspaper title and year should
            be included in the returned dict for later aggregation. Defaults to False.

    Returns:
        dict[str, int]: Dict listing the counts for this issue, ready to be aggregated.
    """
    counts = (
        {
            "np_id": issue["id"].split("-")[0],
            "year": issue["id"].split("-")[1],
        }
        if include_np_yr
        else {}
    )
    counts.update(
        {
            "issues": 1,
            "pages": len(set(issue["pp"])),
            "content_items_out": len(issue["i"]),
            "images": len([item for item in issue["i"] if item["m"]["tp"] == "image"]),
        }
    )
    return counts


def counts_for_rebuilt(
    rebuilt_ci: dict[str, Any], include_np: bool = False, passim: bool = False
) -> dict[str, Union[int, str]]:
    """Define the counts for 1 given rebuilt content-item to match the count keys.

    Args:
        rebuilt_ci (dict[str, Any]): Rebuilt content-item from which to extract counts.
        include_np (bool, optional): Whether to include the title in resulting dict,
            not necessary for on-the-fly computation. Defaults to False.
        passim (bool, optional): True if rebuilt is in passim format. Defaults to False.

    Returns:
        dict[str, Union[int, str]]: Dict with rebuilt (passim) keys and counts for 1 CI.
    """
    split_id = rebuilt_ci["id"].split("-")
    counts = {"np_id": split_id[0]} if include_np else {}
    counts.update(
        {
            "year": split_id[1],
            "issues": "-".join(split_id[:-1]),  # count the issues represented
            "content_items_out": 1,
        }
    )
    if not passim:
        counts.update(
            {
                "ft_tokens": (
                    len(rebuilt_ci["ft"].split()) if "ft" in rebuilt_ci else 0
                ),  # split on spaces to count tokens
            }
        )

    return counts


def compute_stats_in_canonical_bag(
    s3_canonical_issues: db.core.Bag, client: Client | None = None
) -> list[dict[str, Any]]:
    """Computes number of issues and pages per newspaper from a Dask bag of canonical data.

    Args:
        s3_canonical_issues (db.core.Bag): Bag with the contents of canonical files to
            compute statistics on.

    Returns:
        list[dict[str, Any]]: List of counts that match canonical DataStatistics keys.
    """

    print("Fetched all issues, gathering desired information.")
    logger.info("Fetched all issues, gathering desired information.")
    count_df = (
        s3_canonical_issues.map(
            lambda i: counts_for_canonical_issue(i, include_np_yr=True)
        )
        .to_dataframe(
            meta={
                "np_id": str,
                "year": str,
                "issues": int,
                "pages": int,
                "images": int,
                "content_items_out": int,
            }
        )
        .persist()
    )

    # cum the counts for all values collected
    aggregated_df = (
        count_df.groupby(by=["np_id", "year"])
        .agg(
            {
                "issues": sum,
                "pages": sum,
                "content_items_out": sum,
                "images": sum,
            }
        )
        .reset_index()
    ).persist()

    if client is not None:
        # only add the progress bar if the client is defined
        progress(aggregated_df)

    print("Finished grouping and aggregating stats by title and year.")
    logger.info("Finished grouping and aggregating stats by title and year.")
    # return as a list of dicts
    return aggregated_df.to_bag(format="dict").compute()


### DEFINITION of tunique ###


# define locally the nunique() aggregation function for dask
def chunk(s):
    """The function applied to the individual partition (map).
    Part of the ggregating function(s) implementing np.nunique()
    """
    return s.apply(lambda x: list(set(x)))


def agg(s):
    """The function which will aggregate the result from all the partitions (reduce).
    Part of the ggregating function(s) implementing np.nunique()
    """
    s = s._selected_obj
    return s.groupby(level=list(range(s.index.nlevels))).sum()


def finalize(s):
    """The optional function that will be applied to the result of the agg_tu functions.
    Part of the ggregating function(s) implementing np.nunique()
    """
    return s.apply(lambda x: len(set(x)))


# aggregating function implementing np.nunique()
tunique = dd.Aggregation("tunique", chunk, agg, finalize)

### DEFINITION of tunique ###


def compute_stats_in_rebuilt_bag(
    rebuilt_articles: db.core.Bag,
    key: str = "",
    include_np: bool = False,
    passim: bool = False,
    client: Client | None = None,
) -> list[dict[str, Union[int, str]]]:
    """Compute stats on a dask bag of rebuilt output content-items.

    Args:
        rebuilt_articles (db.core.Bag): Bag with the contents of rebuilt files.
        key (str, optional): Optionally title-year pair for on-the-fly computation.
            Defaults to "".
        include_np (bool, optional): Whether to include the title in the groupby,
            not necessary for on-the-fly computation. Defaults to False.
        passim (bool, optional): True if rebuilt is in passim format. Defaults to False.
        client (Client | None, optional): Dask client. Defaults to None.

    Returns:
        list[dict[str, Union[int, str]]]: List of counts that match rebuilt or paassim
            DataStatistics keys.
    """
    # when called in the rebuilt, all the rebuilt articles in the bag
    # are from the same newspaper and year
    print("Fetched all files, gathering desired information.")
    logger.info("Fetched all files, gathering desired information.")

    # define the list of columns in the dataframe
    df_meta = {"np_id": str} if include_np else {}
    df_meta.update(
        {
            "year": str,
            "issues": str,
            "content_items_out": int,
        }
    )
    if not passim:
        df_meta.update(
            {
                "ft_tokens": int,
            }
        )

    rebuilt_count_df = (
        rebuilt_articles.map(
            lambda rf: counts_for_rebuilt(rf, include_np=include_np, passim=passim)
        )
        .to_dataframe(meta=df_meta)
        .persist()
    )

    gp_key = ["np_id", "year"] if include_np else "year"
    # agggregate them at the scale of the entire corpus
    # first groupby title, year and issue to also count the individual issues present
    if not passim:
        aggregated_df = rebuilt_count_df.groupby(by=gp_key).agg(
            {"issues": tunique, "content_items_out": sum, "ft_tokens": sum}
        )
    else:
        aggregated_df = rebuilt_count_df.groupby(by=gp_key).agg(
            {"issues": tunique, "content_items_out": sum}
        )

    # when titles are included, multiple titles and years will be represented
    if include_np:
        aggregated_df = aggregated_df.reset_index().persist()

    msg = "Obtaining the yearly rebuilt statistics"
    if key != "":
        logger.info("%s for %s", msg, key)
    else:
        logger.info(msg)

    if client is not None:
        # only add the progress bar if the client is defined
        progress(aggregated_df)

    return aggregated_df.to_bag(format="dict").compute()


def compute_stats_in_entities_bag(
    s3_entities: db.core.Bag, client: Client | None = None
) -> list[dict[str, Any]]:
    """Compute stats on a dask bag of entities output content-items.

    Args:
        s3_entities (db.core.Bag): Bag with the contents of entity files.
        client (Client | None, optional): Dask client. Defaults to None.

    Returns:
        list[dict[str, Any]]: List of counts that match NE DataStatistics keys.
    """
    count_df = (
        s3_entities.map(
            lambda ci: {
                "np_id": ci["id"].split("-")[0],
                "year": ci["id"].split("-")[1],
                "issues": "-".join(ci["id"].split("-")[:-1]),
                "content_items_out": 1,
                "ne_mentions": len(ci["nes"]),
                "ne_entities": sorted(
                    list(
                        set(
                            [
                                m["wkd_id"]
                                for m in ci["nes"]
                                if "wkd_id" in m and m["wkd_id"] not in ["NIL", None]
                            ]
                        )
                    )
                ),  # sorted list to ensure all are the same
            }
        ).to_dataframe(
            meta={
                "np_id": str,
                "year": str,
                "issues": str,
                "content_items_out": int,
                "ne_mentions": int,
                "ne_entities": object,
            }
        )
        # .explode("ne_entities")
        # .persist()
    )

    count_df["ne_entities"] = count_df["ne_entities"].apply(
        lambda x: x if isinstance(x, list) else [x]
    )
    count_df = count_df.explode("ne_entities").persist()

    # cum the counts for all values collected
    aggregated_df = (
        count_df.groupby(by=["np_id", "year"])
        .agg(
            {
                "issues": tunique,
                "content_items_out": sum,
                "ne_mentions": sum,
                "ne_entities": tunique,
            }
        )
        .reset_index()
    ).persist()

    print("Finished grouping and aggregating stats by title and year.")
    logger.info("Finished grouping and aggregating stats by title and year.")

    if client is not None:
        # only add the progress bar if the client is defined
        progress(aggregated_df)

    # return as a list of dicts
    return aggregated_df.to_bag(format="dict").compute()


def compute_stats_in_langident_bag(
    s3_langident: db.core.Bag, client: Client | None = None
) -> list[dict[str, Any]]:
    """Compute stats on a dask bag of langident output content-items.

    Args:
        s3_langident (db.core.Bag): Bag of lang-id content-items.
        client (Client | None, optional): Dask client. Defaults to None.

    Returns:
        list[dict[str, Any]]:  List of counts that match langident DataStatistics keys.
    """

    def freq(x, col="lang_fd"):
        x[col] = dict(Counter(literal_eval(x[col])))
        return x

    count_df = (
        s3_langident.map(
            lambda ci: {
                "np_id": ci["id"].split("-")[0],
                "year": ci["id"].split("-")[1],
                "issues": "-".join(ci["id"].split("-")[:-1]),
                "content_items_out": 1,
                "images": 1 if ci["tp"] == "img" else 0,
                "lang_fd": "None" if ci["lg"] is None else ci["lg"],
            }
        )
        .to_dataframe(
            meta={
                "np_id": str,
                "year": str,
                "issues": str,
                "content_items_out": int,
                "images": int,
                "lang_fd": object,
            }
        )
        .persist()
    )

    # cum the counts for all values collected
    aggregated_df = (
        count_df.groupby(by=["np_id", "year"])
        .agg(
            {
                "issues": tunique,
                "content_items_out": sum,
                "images": sum,
                "lang_fd": list,
            }
        )
        .reset_index()
    ).persist()

    # Dask dataframes did not support using literal_eval
    agg_bag = aggregated_df.to_bag(format="dict").map(freq)

    if client is not None:
        # only add the progress bar if the client is defined
        progress(agg_bag)

    return agg_bag.compute()


def compute_stats_in_solr_text_bag(
    s3_solr_text: db.core.Bag, client: Client | None = None
) -> list[dict[str, Any]]:
    """Compute stats on a dask bag of content-items formatted for Solr input.

    Args:
        s3_solr_text (db.core.Bag): Bag or Solr formatted content-items.
        client (Client | None, optional): Dask client. Defaults to None.

    Returns:
        list[dict[str, Any]]:  List of counts that match solr text DataStatistics keys.
    """
    count_df = (
        s3_solr_text.map(
            lambda ci: {
                "np_id": ci["meta_journal_s"],
                "year": ci["meta_year_i"],
                "issues": ci["meta_issue_id_s"],
                "content_items_out": 1,
            }
        )
        .to_dataframe(
            meta={
                "np_id": str,
                "year": str,
                "issues": str,
                "content_items_out": int,
            }
        )
        .persist()
    )

    # cum the counts for all values collected
    aggregated_df = (
        count_df.groupby(by=["np_id", "year"])
        .agg(
            {
                "issues": tunique,
                "content_items_out": sum,
            }
        )
        .reset_index()
    ).persist()

    print("Finished grouping and aggregating stats by title and year.")
    logger.info("Finished grouping and aggregating stats by title and year.")

    if client is not None:
        # only add the progress bar if the client is defined
        progress(aggregated_df)

    # return as a list of dicts
    return aggregated_df.to_bag(format="dict").compute()


################ MANIFEST DIFFS #################


def manifest_summary(mnf_json: dict[str, Any], extended_summary: bool = False) -> None:
    """
    Generate a summary of the manifest data.

    Args:
        mnf_json (dict): A dictionary containing manifest data.
        extended_summary (bool, optional): Whether to include extended summary
        with year statistics. Defaults to False.

    Returns:
        None

    Prints: Summary of the manifest including the number of media items, additions,
    and modifications.

    Example:
        >>> manifest_summary(manifest_json)
        Summary of manifest /path/to/manifest.json:
        Number of media items: 10 (8 from set)
        Number of addition at title level: 5
        Number of addition at year level: 3
        Number of modification at title level: 2
        Number of modification at year level: 1
    """
    nb_media_items = len(mnf_json["media_list"])
    nb_addition_title_level = 0
    nb_addition_year_level = 0
    nb_modification_title_level = 0
    nb_modification_year_level = 0

    media_item_set = set()
    title_nbyear = {}

    for media_item in mnf_json["media_list"]:
        media_item_set.add(media_item["media_title"])
        if media_item["update_type"] == "addition":
            if media_item["update_level"] == "title":
                nb_addition_title_level += 1
            elif media_item["update_level"] == "year":
                nb_addition_year_level += 1
        elif media_item["update_type"] == "modification":
            if media_item["update_level"] == "title":
                nb_modification_title_level += 1
            elif media_item["update_level"] == "year":
                nb_modification_year_level += 1
        if extended_summary:
            title_nbyear[media_item["media_title"]] = len(
                [
                    year_stats
                    for year_stats in media_item["media_statistics"]
                    if year_stats["granularity"] == "year"
                ]
            )

    # print summary
    print(
        f"\n*** Summary of manifest: [{mnf_json['mft_s3_path']}] (regardless of any "
        f"modification date):"
    )
    print(f"- Number of media items: {nb_media_items} ({len(media_item_set)} from set)")
    print(f"- Number of addition at title level: {nb_addition_title_level}")
    print(f"- Number of addition at year level: {nb_addition_year_level}")
    print(f"- Number of modification at title level: {nb_modification_title_level}")
    print(f"- Number of modification at year level: {nb_modification_year_level}")
    print(f"- List of media titles:\n{get_media_titles(mnf_json['media_list'])}\n")

    if extended_summary:
        print(
            f"\nExtended summary - Number of years per title "
            f"(regardless of modification/addition or not):"
            f"{nb_modification_year_level}"
        )
        for key, val in title_nbyear.items():
            print(f"- {key:<18}: {val:>5}y")
            print("\n")


def filter_new_or_modified_media(
    rebuilt_mft_json: dict[str, Any], previous_mft_json: dict[str, Any]
) -> dict[str, Any]:
    """
    Compares two manifests to determine new or modified media items.

    Typical use-case is during an atomic update, when only media items added or modified
    compared to the previous process need to be ingested or processed.

    Args:
        rebuilt_mft_json (dict[str, Any]): json of the rebuilt manifest (new).
        previous_mft_json (dict[str, Any]): json of the previous process manifest.

    Returns:
        list[dict[str, Any]]: A manifest identical to 'rebuilt_mft_path' but only with
        media items that are new or modified in the media list.

    Example: >>> new_or_modified = get_new_or_modified_media("new_manifest.json",
    "previous_manifest.json") >>> print(new_or_modified) [{'media_title':
    'new_media_item_1', 'last_modif_date': '2024-04-04T12:00:00Z', etc.},
    {'media_title': 'modified_media_item_2', 'last_modif_date':
    '2024-04-03T12:00:00Z', etc.}]
    """
    filtered_manifest = copy.deepcopy(rebuilt_mft_json)

    # Extract last modification date of each media item of the previous process
    previous_media_items = {
        media["media_title"]: strptime(
            media["last_modification_date"], "%Y-%m-%d %H:%M:%S"
        )
        for media in previous_mft_json["media_list"]
    }

    # Print rebuilt manifest summary
    manifest_summary(rebuilt_mft_json, extended_summary=False)

    # Filter: keep only media items newly added or modified after last process
    filtered_media_list = []
    for rebuilt_media_item in rebuilt_mft_json["media_list"]:
        if rebuilt_media_item["media_title"] not in previous_media_items:
            filtered_media_list.append(rebuilt_media_item)
        elif (
            strptime(rebuilt_media_item["last_modification_date"], "%Y-%m-%d %H:%M:%S")
            > previous_media_items[rebuilt_media_item["media_title"]]
        ):
            filtered_media_list.append(rebuilt_media_item)

    logger.info(
        "\n*** Getting new or modified items:"
        "\nInput (rebuilt) manifest has %s media items.",
        len(get_media_titles(rebuilt_mft_json)),
    )
    logger.info(
        "Resulting filtered manifest has %s media items.", len(filtered_media_list)
    )
    logger.info(
        "Media items that will be newly processed:\n %s\n",
        get_media_titles(filtered_media_list),
    )

    filtered_manifest["media_list"] = filtered_media_list

    return filtered_manifest


def get_media_titles(
    input_data: Union[dict[str, Any], list[dict[str, Any]]]
) -> list[str]:
    """
    Extracts media titles from the input data which can be either a manifest
    or a media list.

    Args:
        input_data (Union[dict[str, Any], list[dict[str, Any]]]): A manifest dictionary
            or the media list of a manifest.

    Returns:
        list[str]: A list of media titles extracted from the input data.
        Ex:  ['Title 1', 'Title 2']
    Raises:
        TypeError: If the input data is not in the expected format.
        KeyError: If the 'media_title' key is not found in the input data.
    """
    if isinstance(input_data, list):
        titles = [media_item["media_title"] for media_item in input_data]
    else:
        titles = [media_item["media_title"] for media_item in input_data["media_list"]]
    return titles


def get_media_item_years(mnf_json: dict[str, Any]) -> dict[str, dict[str, float]]:
    """
    Retrieves the s3 key and size in MB of each year of media items from a manifest.

    Args:
        mnf_json (dict): A manifest dictionary.

    Returns:
       media_items_years (dict): A dictionary where media titles are keys,
            and each value is a dictionary with s3 key as key and its size as value.
    """

    bucket_name = mnf_json["mft_s3_path"].rsplit("/", 1)[0]
    media_items_years = {}

    logger.info("*** Retrieving size info for each key")
    for media_item in tqdm(mnf_json["media_list"]):
        title = media_item["media_title"]
        years = {}

        for media_year in media_item["media_statistics"]:
            if media_year["granularity"] != "year":
                continue

            year_key = title + "/" + media_year["element"] + ".jsonl.bz2"
            s3_key = bucket_name + "/" + year_key
            year_size_b = get_s3_object_size(bucket_name.split("//")[1], year_key)
            year_size_m = (
                round(bytes_to(year_size_b, "m"), 2)
                if year_size_b is not None
                else None
            )
            # print(f"Size in b: {year_size_b}, in mb: {year_size_m}")

            years[s3_key] = year_size_m

        media_items_years[title] = years

    logger.info("*** About the collection if s3 keys for each year:")
    for t, y in media_items_years.items():
        no_s3_keys = [s3k for s3k, size in y.items() if size is None]
        valid_s3_keys = [s3k for s3k, size in y.items() if size is not None]
        logger.info(
            "\t[%s] has %s existing s3 keys and %s missing keys.",
            t,
            len(valid_s3_keys),
            len(no_s3_keys),
        )

    return media_items_years


def remove_media_in_manifest(mnf_json: dict[str, Any], white_list: list[str]) -> None:
    """
    Removes media items from the given manifest JSON object based on a whitelist.
    Typical use case is ingestion or processing only part of the media for whatever reason.

    Parameters:
        mnf_json (dict[str, Any]): The manifest JSON object containing a 'media_list'.
        white_list (list[str]): A list of media titles to be retained in the manifest.

    Returns:
        None: Modifies the input manifest JSON object in-place by removing media items
        not in the whitelist.
    """
    new_media_list = [
        media_item
        for media_item in mnf_json["media_list"]
        if media_item["media_title"] in white_list
    ]
    mnf_json["media_list"] = new_media_list



"""Reusable functions to read/write data from/to our S3 drive.
Warning: 2 boto libraries are used, and need to be kept until third party lib dependencies are solved.

"""

import os
import logging
import json
import warnings
import bz2
from typing import Union
import boto3
from smart_open.s3 import iter_bucket
from smart_open import open as s_open
from dotenv import load_dotenv
import botocore

from impresso_commons.utils import _get_cores

logger = logging.getLogger(__name__)


def get_storage_options():
    # load environment variables from local .env files
    load_dotenv()
    return {
        "client_kwargs": {"endpoint_url": "https://os.zhdk.cloud.switch.ch"},
        "key": os.environ["SE_ACCESS_KEY"],
        "secret": os.environ["SE_SECRET_KEY"],
    }


IMPRESSO_STORAGEOPT = get_storage_options()

_WARNING = """As the boto library is being removed from this package the 
    following functions are depreciated and should be replaced:
    - get_s3_connection, to be replaced by get_s3_resource.
    - get_bucket_boto3 and get_bucket, to be replaced by get_or_create_bucket.
    
    The objects returned by boto3 differ from the ones returned by boto, which 
    might cause issues in the code. Please refer to this migration guide:
    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/migrations3.html
"""
_WARNED = False


def get_s3_client(host_url="https://os.zhdk.cloud.switch.ch/"):

    # load environment variables from local .env files
    load_dotenv()
    if host_url is None:
        try:
            host_url = os.environ["SE_HOST_URL"]
        except Exception:
            raise

    try:
        access_key = os.environ["SE_ACCESS_KEY"]
    except Exception:
        raise

    try:
        secret_key = os.environ["SE_SECRET_KEY"]
    except Exception:
        raise

    return boto3.client(
        "s3",
        aws_secret_access_key=secret_key,
        aws_access_key_id=access_key,
        endpoint_url=host_url,
    )


def get_s3_resource(host_url="https://os.zhdk.cloud.switch.ch/"):
    """Get a boto3 resource object related to an S3 drive.

    Assumes that two environment variables are set:
    `SE_ACCESS_KEY` and `SE_SECRET_KEY`.

    :param host_url: the s3 endpoint's URL
    :type host_url: string
    :rtype: `boto3.resources.factory.s3.ServiceResource`
    """

    # load environment variables from local .env files
    load_dotenv()
    if host_url is None:
        try:
            host_url = os.environ["SE_HOST_URL"]
        except Exception:
            raise

    try:
        access_key = os.environ["SE_ACCESS_KEY"]
    except Exception:
        raise

    try:
        secret_key = os.environ["SE_SECRET_KEY"]
    except Exception:
        raise

    return boto3.resource(
        "s3",
        aws_secret_access_key=secret_key,
        aws_access_key_id=access_key,
        endpoint_url=host_url,
    )


def get_s3_connection(host="os.zhdk.cloud.switch.ch"):
    """Create a boto connection to impresso's S3 drive.

    Assumes that two environment variables are set: `SE_ACCESS_KEY` and
    `SE_SECRET_KEY`.

    Note:
        This function is depreciated, as it used boto instead of boto3,
        please prioritize using get_s3_resource instead.

    :param host_url: the s3 endpoint's URL
    :type host_url: string
    :rtype: `boto3.resources.factory.s3.ServiceResource`
    """
    global _WARNED
    if not _WARNED:
        logger.warning(_WARNING)
        warnings.warn(_WARNING, DeprecationWarning)
        _WARNED = True
    return get_s3_resource()


def get_or_create_bucket(name, create=False, versioning=True):
    """Create a boto3 s3 connection and returns the requested bucket.

    It is possible to ask for creating a new bucket
    with the specified name (in case it does not exist), and (optionally)
    to turn on the versioning on the newly created bucket.
    >>> b = get_bucket('testb', create=False)
    >>> b = get_bucket('testb', create=True)
    >>> b = get_bucket('testb', create=True, versioning=False)

    :param name: the bucket's name
    :type name: string
    :param create: creates the bucket if not yet existing
    :type create: boolean
    :param versioning: whether the new bucket should be versioned
    :type versioning: boolean
    :return: an s3 bucket
    :rtype: `boto3.resources.factory.s3.Bucket`
    """
    s3r = get_s3_resource()
    # try to fetch the specified bucket -- may return an empty list
    bucket = [b for b in s3r.buckets.all() if b.name == name]

    try:
        assert len(bucket) > 0
        return bucket[0]

    # bucket not found
    except AssertionError:
        if create:
            bucket = s3r.create_bucket(Bucket=name)
            print(f"New bucket {name} was created")
        else:
            print(f"Bucket {name} not found")
            return None

    # enable versioning
    if versioning:
        bucket_versioning = s3r.BucketVersioning(name)
        bucket_versioning.enable()

    print(f"Versioning: {bucket_versioning.status}")

    return bucket


def get_bucket(name, create=False, versioning=True):
    """Create a boto s3 connection and returns the requested bucket.

    It is possible to ask for creating a new bucket
    with the specified name (in case it does not exist), and (optionally)
    to turn on the versioning on the newly created bucket.
    >>> b = get_bucket('testb', create=False)
    >>> b = get_bucket('testb', create=True)
    >>> b = get_bucket('testb', create=True, versioning=False)

    Note:
        This function is depreciated, please prioritize using
        get_or_create_bucket or get_boto3_bucket instead.

    :param name: the bucket's name
    :type name: string
    :param create: creates the bucket if not yet existing
    :type create: boolean
    :param versioning: whether the new bucket should be versioned
    :type versioning: boolean
    :return: an s3 bucket
    :rtype: `boto3.resources.factory.s3.Bucket`
    """
    global _WARNED
    if not _WARNED:
        logger.warning(_WARNING)
        warnings.warn(_WARNING, DeprecationWarning)
        _WARNED = True
    return get_or_create_bucket(name, create=create, versioning=versioning)


def get_bucket_boto3(name, create=False, versioning=True):
    """Get a boto3 s3 resource and returns the requested bucket.

    It is possible to ask for creating a new bucket
    with the specified name (in case it does not exist), and (optionally)
    to turn on the versioning on the newly created bucket.

    >>> b = get_bucket('testb', create=False)
    >>> b = get_bucket('testb', create=True)
    >>> b = get_bucket('testb', create=True, versioning=False)

    Note:
        This function is depreciated, please prioritize using
        get_or_create_bucket or get_boto3_bucket instead.

    :param name: the bucket's name
    :type name: string
    :param create: creates the bucket if not yet existing
    :type create: boolean
    :param versioning: whether the new bucket should be versioned
    :type versioning: boolean
    :return: an s3 bucket
    :rtype: `boto3.resources.factory.s3.Bucket`
    """
    global _WARNED
    if not _WARNED:
        logger.warning(_WARNING)
        warnings.warn(_WARNING, DeprecationWarning)
        _WARNED = True
    return get_or_create_bucket(name, create=create, versioning=versioning)


def s3_get_articles(issue, bucket, workers=None):
    """Read a newspaper issue from S3 and return the articles it contains.

    NB: Content items with type = "ad" (advertisement) are filtered out.

    :param issue: the newspaper issue
    :type issue: an instance of `impresso_commons.path.IssueDir`
    :param bucket: the input s3 bucket
    :type bucket: `boto.s3.bucket.Bucket`
    :param workers: number of workers for the iter_bucket function.
        If None, will be the number of detected CPUs.
    :return: a list of articles (dictionaries)
    """
    nb_workers = _get_cores() if workers is None else workers
    issue_data = list(iter_bucket(bucket, prefix=issue.path, workers=nb_workers))
    print(issue_data)
    issue_data = issue_data[0][1]
    issue_json = json.loads(issue_data.decode("utf-8"))
    articles = [item for item in issue_json["i"] if item["m"]["tp"] == "article"]
    return articles


def s3_get_pages(issue_id, page_names, bucket):
    """Read in canonical text data for all pages in a given newspaper issue.

    :param issue_id: the canonical issue id (e.g. "IMP-1990-03-15-a")
    :type issue_id: string
    :param page_names: a list of canonical page filenames
        (e.g. "IMP-1990-03-15-a-p0001.json")
    :type page_names: list of strings
    :param bucket: the s3 bucket where the pages to be read are stored
    :type bucket: instance of `boto.Bucket`
    :return: a dictionary with page filenames as keys, and JSON data as values.
    """
    pages = {}

    for page in page_names.values():
        key_name = os.path.join(issue_id.replace("-", "/"), page)
        key = bucket.get_key(key_name, validate=False)
        logger.info(f"reading page {key_name}")
        content = key.get_contents_as_string()
        pages[key.name.split("/")[-1]] = json.loads(content.decode("utf-8"))
    return pages
    """
    return {
        key.name.split('/')[-1]: json.loads(content.decode('utf-8'))
        for key, content in s3_iter_bucket(
            bucket,
            prefix=issue_id.replace('-', '/')
        )
        if key.name.split('/')[-1] in list(page_names.values())
    }
    """


def get_s3_versions(bucket_name, key_name):
    """Get versioning information for a given key.

    **NB:** it assumes a versioned bucket.

    :param bucket_name: the bucket's name
    :type bucket_name: string
    :param key_name: the key's name
    :type key_name: string
    :return: for each version, the version id and the last modified date
    :rtype: a list of tuples, where tuple[0] is a string and tuple[1] a
        `datetime` instance.
    """

    client = get_s3_resource()

    # may be worth comparing with
    # client.list_object_versions(prefix)
    versions = client.Bucket(bucket_name).object_versions.filter(Prefix=key_name)

    version_ids = [
        (v.get().get("VersionId"), v.get().get("LastModified"))
        for v in versions
        if v.size is not None
    ]
    return version_ids


def get_s3_versions_client(client, bucket_name, key_name):

    versions = client.Bucket(bucket_name).object_versions.filter(Prefix=key_name)

    version_ids = [
        (v.get().get("VersionId"), v.get().get("LastModified"))
        for v in versions
        if v.size is not None
    ]
    return version_ids


def read_jsonlines(key_name, bucket_name):
    """
    Given an S3 key pointing to a jsonl.bz2 archives, extracts and returns lines (=one json doc per line).
    Usage example:
    >>> lines = db.from_sequence(read_jsonlines(s3r, key_name , bucket_name))
    >>> print(lines.count().compute())
    >>> lines.map(json.loads).pluck('ft').take(10)
    :param bucket_name: name of bucket
    :type bucket_name: str
    :param key_name: name of key, without S3 prefix
    :type key_name: str
    :return:
    """
    s3r = get_s3_resource()
    body = s3r.Object(bucket_name, key_name).get()["Body"]
    data = body.read()
    text = bz2.decompress(data).decode("utf-8")
    for line in text.split("\n"):
        if line != "":
            yield line


def readtext_jsonlines(key_name, bucket_name):
    """
    Given an S3 key pointing to a jsonl.bz2 archives, extracts and returns lines (=one json doc per line)
    with limited textual information, leaving out OCR metadata (box, offsets).
    This can serve as the starting point for pure textual processing (NE, text-reuse, topics)
    Usage example:
    >>> lines = db.from_sequence(readtext_jsonlines(s3r, key_name , bucket_name))
    >>> print(lines.count().compute())
    >>> lines.map(json.loads).pluck('ft').take(10)
    :param bucket_name: name of bucket
    :type bucket_name: str
    :param key_name: name of key, without S3 prefix
    :type key_name: str
    :return: JSON formatted str
    """
    s3r = get_s3_resource()
    body = s3r.Object(bucket_name, key_name).get()["Body"]
    data = body.read()
    text = bz2.decompress(data).decode("utf-8")
    for line in text.split("\n"):
        if line != "":
            article_json = json.loads(line)
            text = article_json["ft"]
            if len(text) != 0:
                article_reduced = {
                    k: article_json[k]
                    for k in article_json
                    if k == "id"
                    or k == "s3v"
                    or k == "ts"
                    or k == "ft"
                    or k == "tp"
                    or k == "pp"
                    or k == "lg"
                    or k == "t"
                }
                yield json.dumps(article_reduced)


def upload(partition_name, newspaper_prefix=None, bucket_name=None):

    if newspaper_prefix is not None:
        key_name = os.path.join("/", newspaper_prefix, partition_name.split("/")[-1])
    else:
        key_name = partition_name.split("/")[-1]
    s3 = get_s3_resource()
    try:
        bucket = s3.Bucket(bucket_name)
        logger.info(bucket.name)
        bucket.upload_file(partition_name, key_name)
        logger.info(f"Uploaded {partition_name} to {key_name}")
        return True, partition_name
    except Exception as e:
        logger.error(e)
        logger.error(f"The upload of {partition_name} failed with error {e}")
        return False, partition_name


def upload_to_s3(local_path: str, path_within_bucket: str, bucket_name: str) -> bool:
    """Upload a file to an S3 bucket.

    Args:
        local_path (str): The local file path to upload.
        path_within_bucket (str): The path within the bucket where the file will be uploaded.
        bucket_name (str): The name of the S3 bucket (without any partitions).

    Returns:
        bool: True if the upload is successful, False otherwise.
    """
    bucket = get_boto3_bucket(bucket_name)
    try:
        # ensure the path within the bucket is only the key
        path_within_bucket = path_within_bucket.replace("s3://", "")
        bucket.upload_file(local_path, path_within_bucket)
        logger.info(f"Uploaded {path_within_bucket} to s3://{bucket_name}.")
        return True
    except Exception as e:
        logger.error(e)
        logger.error(f"The upload of {local_path} failed with error {e}")
        return False


def get_boto3_bucket(bucket_name: str):
    s3 = get_s3_resource()
    return s3.Bucket(bucket_name)


def fixed_s3fs_glob(path: str, boto3_bucket=None):
    """
    From Benoit, impresso-pyimages package
    A custom glob function as the s3fs one seems to be unable to list more than 1000 elements on the switch S3
    :param path:
    :return:
    """
    if boto3_bucket is None:
        if path.startswith("s3://"):
            path = path[len("s3://") :]
        bucket_name = path.split("/")[0]
        base_path = "/".join(path.split("/")[1:])  # Remove bucket name
        boto3_bucket = get_boto3_bucket(bucket_name)
    else:
        bucket_name = boto3_bucket.name
        base_path = path
    base_path, suffix_path = base_path.split("*")
    filenames = [
        "s3://"
        + os.path.join(
            bucket_name, o.key
        )  # prepend bucket-name as it is necessary for s3fs
        for o in boto3_bucket.objects.filter(Prefix=base_path)
        if o.key.endswith(suffix_path)
    ]
    return filenames


def alternative_read_text(
    s3_key: str, s3_credentials: dict, line_by_line: bool = True
) -> Union[list[str], str]:
    """Read from S3 a line-separated text file (e.g. `*.jsonl.bz2`).

    Note:
        The reason for this function is a bug in `dask.bag.read_text()`
        which breaks on buckets having >= 1000 keys.
        It raises a `FileNotFoundError`.
    """
    logger.info(f"reading {s3_key}")
    session = boto3.Session(
        aws_access_key_id=s3_credentials["key"],
        aws_secret_access_key=s3_credentials["secret"],
    )
    s3_endpoint = s3_credentials["client_kwargs"]["endpoint_url"]
    transport_params = {
        "client": session.client("s3", endpoint_url=s3_endpoint),
    }

    if line_by_line:
        with s_open(s3_key, "r", transport_params=transport_params) as infile:
            text = infile.readlines()
    else:
        with s_open(s3_key, "r", transport_params=transport_params) as infile:
            text = infile.read()

    return text


def get_s3_object_size(bucket_name, key):
    """
    Get the size of an object (key) in an S3 bucket.

    Args:
        bucket_name (str): The name of the S3 bucket.
        key (str): The key (object) whose size you want to retrieve.

    Returns:
        int: The size of the object in bytes, or None if the object doesn't exist.
    """
    s3_client = get_s3_client()

    try:
        # Get the object metadata to retrieve its size
        response = s3_client.head_object(Bucket=bucket_name, Key=key)
        size = response["ContentLength"]
        return int(size)
    except botocore.exceptions.ClientError as err:
        logger.error(f"Error: {err} for {key} in {bucket_name}")
        return None



import re
from impresso_commons.path.path_fs import IssueDir
from datetime import date


def parse_canonical_filename(filename):
    """Parse a canonical page names into its components.

    :param filename: the filename to parse
    :type filename: string
    :return: a tuple

    >>> filename = "GDL-1950-01-02-a-i0002"
    >>> parse_canonical_filename(filename)
    >>> ('GDL', ('1950', '01', '02'), 'a', 'i', 2, '')
    """
    regex = re.compile(
        (
            r'^(?P<np>[A-Za-z0-9]+)-(?P<year>\d{4})'
            r'-(?P<month>\d{2})-(?P<day>\d{2})'
            r'-(?P<ed>[a-z])-(?P<type>[p|i])(?P<pgnb>\d{4})(?P<ext>.*)?$'
        )
    )
    result = re.match(regex, filename)
    newspaper_id = result.group("np")
    date = (result.group("year"), result.group("month"), result.group("day"))
    page_number = int(result.group("pgnb"))
    edition = result.group("ed")
    filetype = result.group("type")
    extension = result.group("ext")
    return (
        newspaper_id,
        date,
        edition,
        filetype,
        page_number,
        extension
    )


def id2IssueDir(id, path):
    """
    TODO: documentation
    """
    newspaper, year, month, day, edition = id.split("-")
    year = int(year)
    month = int(month)
    day = int(day)
    return IssueDir(newspaper, date(year, month, day), edition, path)



"""Functions and CLI to rebuild text from impresso's canonical format.
For EPFL members, this script can be scaled by running it using Runai, 
as documented on https://github.com/impresso/impresso-infrastructure/blob/main/howtos/runai.md.

Usage:
    rebuilder.py rebuild_articles --input-bucket=<b> --log-file=<f> --output-dir=<od> --filter-config=<fc> [--format=<fo> --scheduler=<sch> --output-bucket=<ob> --verbose --clear --languages=<lgs> --nworkers=<nw> --git-repo=<gr> --temp-dir=<tp> --prev-manifest=<pm>]

Options:

--input-bucket=<b>  S3 bucket where canonical JSON data will be read from
--output-bucket=<ob>  Rebuilt data will be uploaded to the specified s3 bucket (otherwise no upload)
--log-file=<f>  Path to log file
--scheduler=<sch>  Tell dask to use an existing scheduler (otherwise it'll create one)
--filter-config=<fc>  A JSON configuration file specifying which newspaper issues will be rebuilt
--verbose  Set logging level to DEBUG (by default is INFO)
--clear  Remove output directory before and after rebuilding
--format=<fo>  Rebuilt format to use (can be "solr" or "passim")
--languages=<lgs>  Languages to filter the articles to rebuild on.
--nworkers=<nw>  number of workers for (local) Dask client.
--git-repo=<gr>   Local path to the "impresso-text-acquisition" git directory (including it).
--temp-dir=<tp>  Temporary directory in which to clone the impresso-data-release git repository.
--prev-manifest=<pm>  Optional S3 path to the previous manifest to use for the manifest generation.
"""  # noqa: E501

import sys
import traceback
import datetime
import json
import pathlib
import logging
import os
import shutil
import signal
from typing import Any, Optional
import git

import dask.bag as db
import jsonlines
from dask.distributed import Client, progress
from docopt import docopt
from smart_open import smart_open

from impresso_commons.path import parse_canonical_filename
from impresso_commons.path.path_s3 import read_s3_issues
from impresso_commons.text.helpers import (
    read_issue_pages,
    rejoin_articles,
    reconstruct_iiif_link,
    insert_whitespace,
)
from impresso_commons.utils import Timer, timestamp
from impresso_commons.utils.s3 import get_s3_resource

from impresso_commons.versioning.data_manifest import DataManifest
from impresso_commons.versioning.helpers import compute_stats_in_rebuilt_bag

logger = logging.getLogger(__name__)


TYPE_MAPPINGS = {
    "article": "ar",
    "ar": "ar",
    "advertisement": "ad",
    "ad": "ad",
    "pg": None,
    "image": "img",
    "table": "tb",
    "death_notice": "ob",
    "weather": "w",
}
# TODO KB data: add familial announcement?


def rebuild_text(
    page: list[dict], language: Optional[str], string: Optional[str] = None
) -> tuple[str, dict[list], dict[list]]:
    """Rebuild the text of an article for Solr ingestion.

    If `string` is not `None`, then the rebuilt text is appended to it.

    Args:
        page (list[dict]): Newspaper page conforming to the impresso JSON pages schema.
        language (str | None): Language of the article being rebuilt
        string (str | None, optional): Rebuilt text of previous page. Defaults to None.

    Returns:
        tuple[str, dict[list], dict[list]]: [0] Article fulltext, [1] offsets and
            [2] coordinates of token regions.
    """

    coordinates = {"regions": [], "tokens": []}

    offsets = {"line": [], "para": [], "region": []}

    if string is None:
        string = ""

    # in order to be able to keep line break information
    # we iterate over a list of lists (lines of tokens)
    for region in page:

        if len(string) > 0:
            offsets["region"].append(len(string))

        coordinates["regions"].append(region["c"])

        for para in region["p"]:

            if len(string) > 0:
                offsets["para"].append(len(string))

            for line in para["l"]:

                for n, token in enumerate(line["t"]):
                    region = {}
                    if "c" not in token:
                        print(f"'c' was not present in token: {token}, line: {line}")
                        continue
                    region["c"] = token["c"]
                    region["s"] = len(string)

                    if "hy" in token:
                        region["l"] = len(token["tx"][:-1]) - 1
                        region["hy1"] = True
                    elif "nf" in token:
                        region["l"] = len(token["nf"])
                        region["hy2"] = True

                        token_text = token["nf"] if token["nf"] is not None else ""
                    else:
                        if token["tx"]:
                            region["l"] = len(token["tx"])
                        else:
                            region["l"] = 0

                        token_text = token["tx"] if token["tx"] is not None else ""

                    # don't add the tokens corresponding to the first part of a hyphenated word
                    if "hy" not in token:
                        next_token = (
                            line["t"][n + 1]["tx"] if n != len(line["t"]) - 1 else None
                        )
                        ws = insert_whitespace(
                            token["tx"],
                            next_t=next_token,
                            prev_t=line["t"][n - 1]["tx"] if n != 0 else None,
                            lang=language,
                        )
                        string += f"{token_text} " if ws else f"{token_text}"

                    # if token is the last in a line
                    if n == len(line["t"]) - 1:
                        if "hy" in token:
                            offsets["line"].append(region["s"])
                        else:
                            token_length = len(token["tx"]) if token["tx"] else 0
                            offsets["line"].append(region["s"] + token_length)

                    coordinates["tokens"].append(region)

    return (string, coordinates, offsets)


def rebuild_text_passim(
    page: list[dict], language: Optional[str], string: Optional[str] = None
) -> tuple[str, list[dict]]:
    """The text rebuilding function from pages for passim.

    If `string` is not `None`, then the rebuilt text is appended to it.

    Args:
        page (list[dict]): Newspaper page conforming to the impresso JSON pages schema.
        language (str | None): Language of the article being rebuilt
        string (str | None, optional): Rebuilt text of previous page. Defaults to None.

    Returns:
        tuple[str, list[dict]]: [0] article fulltext, [1] coordinates of token regions.
    """

    regions = []

    if string is None:
        string = ""

    # in order to be able to keep line break information
    # we iterate over a list of lists (lines of tokens)

    for region in page:

        for para in region["p"]:

            for line in para["l"]:

                for n, token in enumerate(line["t"]):

                    region_string = ""

                    if "c" not in token:
                        # if the coordniates are missing, they should be skipped
                        logger.debug("Missing 'c' in token %s", token)
                        print(f"Missing 'c' in token {token}")
                        continue

                    # each page region is a token
                    output_region = {
                        "start": None,
                        "length": None,
                        "coords": {
                            "x": token["c"][0],
                            "y": token["c"][1],
                            "w": token["c"][2],
                            "h": token["c"][3],
                        },
                    }

                    if len(string) == 0:
                        output_region["start"] = 0
                    else:
                        output_region["start"] = len(string)

                    # if token is the last in a line
                    if n == len(line["t"]) - 1:
                        tmp = f"{token['tx']}\n"
                        region_string += tmp
                    else:
                        ws = insert_whitespace(
                            token["tx"],
                            next_t=line["t"][n + 1]["tx"],
                            prev_t=line["t"][n - 1]["tx"] if n != 0 else None,
                            lang=language,
                        )
                        region_string += f"{token['tx']} " if ws else f"{token['tx']}"

                    string += region_string
                    output_region["length"] = len(region_string)
                    regions.append(output_region)

    return (string, regions)


def rebuild_for_solr(content_item: dict[str, Any]) -> dict[str, Any]:
    """Rebuilds the text of an article content-item given its metadata as input.

    Note:
        This rebuild function is thought especially for ingesting the newspaper
        data into our Solr index.

    Args:
        content_item (dict[str, Any]): The content-item to rebuilt using its metadata.

    Returns:
        dict[str, Any]: The rebuilt content-item following the Impresso JSON Schema.
    """
    t = Timer()

    article_id = content_item["m"]["id"]
    logger.debug("Started rebuilding article %s", article_id)

    issue_id = "-".join(article_id.split("-")[:-1])

    page_file_names = {
        p: f"{issue_id}-p{str(p).zfill(4)}.json" for p in content_item["m"]["pp"]
    }

    year, month, day, _, ci_num = article_id.split("-")[1:]
    d = datetime.date(int(year), int(month), int(day))

    raw_type = content_item["m"]["tp"]

    if raw_type in TYPE_MAPPINGS:
        mapped_type = TYPE_MAPPINGS[raw_type]
    else:
        mapped_type = raw_type

    fulltext = ""
    linebreaks = []
    parabreaks = []
    regionbreaks = []

    article_lang = content_item["m"]["l"] if "l" in content_item["m"] else None

    # if the reading order is not defined, use the number associated to each CI
    reading_order = (
        content_item["m"]["ro"] if "ro" in content_item["m"] else int(ci_num[1:])
    )

    article = {
        "id": article_id,
        "pp": content_item["m"]["pp"],
        "d": d.isoformat(),
        "olr": False if mapped_type is None else True,
        "ts": timestamp(),
        "lg": article_lang,
        "tp": mapped_type,
        "ro": reading_order,
        "s3v": content_item["m"]["s3v"] if "s3v" in content_item["m"] else None,
        "ppreb": [],
        "lb": [],
        "cc": content_item["m"]["cc"],
    }

    if mapped_type == "img":
        article["iiif_link"] = reconstruct_iiif_link(content_item)

    if "t" in content_item["m"]:
        article["t"] = content_item["m"]["t"]

    if mapped_type != "img":
        for n, page_no in enumerate(article["pp"]):

            page = content_item["pprr"][n]

            if fulltext == "":
                fulltext, coords, offsets = rebuild_text(page, article_lang)
            else:
                fulltext, coords, offsets = rebuild_text(page, article_lang, fulltext)

            linebreaks += offsets["line"]
            parabreaks += offsets["para"]
            regionbreaks += offsets["region"]

            page_doc = {
                "id": page_file_names[page_no].replace(".json", ""),
                "n": page_no,
                "t": coords["tokens"],
                "r": coords["regions"],
            }
            article["ppreb"].append(page_doc)
        article["lb"] = linebreaks
        article["pb"] = parabreaks
        article["rb"] = regionbreaks
        logger.debug("Done rebuilding article %s (Took %s)", article_id, t.stop())
        article["ft"] = fulltext

    return article


def rebuild_for_passim(content_item: dict[str, Any]) -> dict[str, Any]:
    """Rebuilds the text of an article content-item to be used with passim.

    Args:
        content_item (dict[str, Any]): The content-item to rebuilt using its metadata.

    Returns:
        dict[str, Any]: The rebuilt content-item built for passim.
    """
    np, date, _, _, _, _ = parse_canonical_filename(content_item["m"]["id"])

    article_id = content_item["m"]["id"]
    logger.debug("Started rebuilding article %s", article_id)
    issue_id = "-".join(article_id.split("-")[:-1])

    page_file_names = {
        p: f"{issue_id}-p{str(p).zfill(4)}.json" for p in content_item["m"]["pp"]
    }

    article_lang = content_item["m"]["l"] if "l" in content_item["m"] else None

    passim_document = {
        "series": np,
        "date": f"{date[0]}-{date[1]}-{date[2]}",
        "id": content_item["m"]["id"],
        "cc": content_item["m"]["cc"],
        "lg": article_lang,
        "pages": [],
    }

    if "t" in content_item["m"]:
        passim_document["title"] = content_item["m"]["t"]

    fulltext = ""
    for n, page_no in enumerate(content_item["m"]["pp"]):

        page = content_item["pprr"][n]

        if fulltext == "":
            fulltext, regions = rebuild_text_passim(page, article_lang)
        else:
            fulltext, regions = rebuild_text_passim(page, article_lang, fulltext)

        page_doc = {
            "id": page_file_names[page_no].replace(".json", ""),
            "seq": page_no,
            "regions": regions,
        }
        passim_document["pages"].append(page_doc)

    passim_document["text"] = fulltext

    return passim_document


def compress(key, json_files, output_dir):
    """Merges a set of JSON line files into a single compressed archive.

    :param key: signature of the newspaper issue (e.g. GDL-1900)
    :type key: str
    :param json_files: input JSON line files
    :type json_files: list
    :param output_dir: directory where to write the output file
    :type outp_dir: str
    :return: a tuple with: sorting key [0] and path to serialized file [1].
    :rytpe: tuple

    .. note::

        `sort_key` is expected to be the concatenation of newspaper ID and year
        (e.g. GDL-1900).
    """

    newspaper, year = key.split("-")
    filename = f"{newspaper}-{year}.jsonl.bz2"
    filepath = os.path.join(output_dir, filename)
    logger.info("Compressing %s JSON files into %s", len(json_files), filepath)
    print(f"Compressing {len(json_files)} JSON files into {filepath}")

    with smart_open(filepath, "wb") as fout:
        writer = jsonlines.Writer(fout)

        for json_file in json_files:
            with open(json_file, "r", encoding="utf-8") as inpf:
                reader = jsonlines.Reader(inpf)
                articles = list(reader)
                writer.write_all(articles)
            logger.info(
                "Written %s docs from %s to %s", len(articles), json_file, filepath
            )

        writer.close()

    for json_file in json_files:
        os.remove(json_file)

    temp_dir = os.path.dirname(json_files[0])
    os.rmdir(temp_dir)
    logger.info("Removed temporary directory and files in %s", temp_dir)

    return (key, filepath)


def upload(sort_key, filepath, bucket_name=None):
    """Uploads a file to a given S3 bucket.

    :param sort_key: the key used to group articles (e.g. "GDL-1900")
    :type sort_key: str
    :param filepath: path of the file to upload to S3
    :type filepath: str
    :param bucket_name: name of S3 bucket where to upload the file
    :type bucket_name: str
    :return: a tuple with [0] whether the upload was successful (boolean) and
        [1] the path of the uploaded file (string)

    .. note::

        `sort_key` is expected to be the concatenation of newspaper ID and year
        (e.g. GDL-1900).
    """
    # create connection with bucket
    # copy contents to s3 key
    newspaper, _ = sort_key.split("-")
    key_name = f"{newspaper}/{os.path.basename(filepath)}"
    if "/" in bucket_name:
        # if the provided bucket also contains a partition, add it to the key name
        bucket_name, partition = bucket_name.split("/")
        key_name = f"{partition}/{key_name}"
    s3 = get_s3_resource()
    try:
        bucket = s3.Bucket(bucket_name)
        bucket.upload_file(filepath, key_name)
        logger.info("Uploaded %s to %s", filepath, key_name)
        return True, filepath
    except Exception as e:
        logger.error(e)
        logger.error("The upload of %s failed with error %s", filepath, e)
        return False, filepath


def cleanup(upload_success, filepath):
    """Removes a file if it has been successfully uploaded to S3.
    :param upload_success: whether the upload was successful
    :type upload_success: bool
    :param filepath: path to the uploaded file
    :type filepath: str
    """
    if upload_success and os.path.exists(filepath):
        try:
            os.remove(filepath)
            logger.info("Removed temporary file %s", filepath)
        except Exception as e:
            logger.warning("Error %s occurred when removing %s", e, filepath)
    else:
        logger.info("Not removing %s as upload has failed", filepath)


def _article_has_problem(article):
    """Helper function to keep articles with problems.

    :param article: input article
    :type article: dict
    :return: `True` or `False`
    :rtype: boolean
    """
    return article["has_problem"]


def _article_without_problem(article):
    """Helper function to keep articles without problems.

    :param article: input article
    :type article: dict
    :return: `True` or `False`
    :rtype: boolean
    """
    if article["has_problem"]:
        logger.warning("Article %s won't be rebuilt.", article["m"]["id"])
    return not article["has_problem"]


def rebuild_issues(
    issues, input_bucket, output_dir, dask_client, _format="solr", filter_language=None
):
    """Rebuild a set of newspaper issues into a given format.

    :param issues: issues to rebuild
    :type issues: list of `IssueDir` objects
    :param input_bucket: name of input s3 bucket
    :type input_bucket: str
    :param outp_dir: local directory where to store the rebuilt files
    :type outp_dir: str
    :return: a list of tuples (see return type of `upload`)
    :rtype: list of tuples
    """

    def mkdir(path):
        if not os.path.exists(path):
            pathlib.Path(path).mkdir(parents=True, exist_ok=True)
        else:
            for f in os.listdir(path):
                os.remove(os.path.join(path, f))

    # determine which rebuild function to apply
    if _format == "solr":
        rebuild_function = rebuild_for_solr
    elif _format == "passim":
        rebuild_function = rebuild_for_passim
    else:
        raise NotImplementedError

    # create a temporary output directory named after newspaper and year
    # e.g. IMP-1994
    issue, _ = issues[0]
    key = f"{issue.journal}-{issue.date.year}"
    issue_dir = os.path.join(output_dir, key)
    mkdir(issue_dir)

    # warning about large graph comes here
    print("Fleshing out articles by issue...")
    issues_bag = db.from_sequence(issues, partition_size=3)

    faulty_issues = (
        issues_bag.filter(lambda i: len(i[1]["pp"]) == 0)
        .map(lambda i: i[1])
        .pluck("id")
        .compute()
    )
    msg = f"Issues with no pages (will be skipped): {faulty_issues}"
    logger.debug(msg)
    print(msg)
    del faulty_issues
    msg = f"Number of partitions: {issues_bag.npartitions}"
    logger.debug(msg)
    print(msg)

    articles_bag = (
        issues_bag.filter(lambda i: len(i[1]["pp"]) > 0)
        .starmap(read_issue_pages, bucket=input_bucket)
        .starmap(rejoin_articles)
        .flatten()
        .persist()
    )

    faulty_articles_n = (
        articles_bag.filter(_article_has_problem).pluck("m").pluck("id").compute()
    )
    msg = f"Skipped articles: {faulty_articles_n}"
    logger.debug(msg)
    print(msg)
    del faulty_articles_n

    articles_bag = (
        articles_bag.filter(_article_without_problem).map(rebuild_function).persist()
    )

    def has_language(ci):
        if "lg" not in ci:
            return False
        return ci["lg"] in filter_language

    if filter_language:
        filtered_articles = articles_bag.filter(has_language).persist()
        print(filtered_articles.count().compute())
        stats_for_issues = compute_stats_in_rebuilt_bag(filtered_articles, key)
        result = filtered_articles.map(json.dumps).to_textfiles(f"{issue_dir}/*.json")
    else:
        stats_for_issues = compute_stats_in_rebuilt_bag(articles_bag, key)
        result = articles_bag.map(json.dumps).to_textfiles(f"{issue_dir}/*.json")

    dask_client.cancel(issues_bag)
    logger.info("done.")
    print("done.")

    return (key, result, stats_for_issues)


def init_logging(level, file):
    """Initialises the root logger.

    :param level: desired level of logging (default: logging.INFO)
    :type level: int
    :param file:
    :type file: str
    :return: the initialised logger
    :rtype: `logging.RootLogger`

    .. note::

        It's basically a duplicate of `impresso_commons.utils.init_logger` but
        I could not get it to work properly, so keeping this duplicate.
    """
    # Initialise the logger
    root_logger = logging.getLogger("")
    root_logger.setLevel(level)

    if file is not None:
        handler = logging.FileHandler(filename=file, mode="w")
    else:
        handler = logging.StreamHandler()

    formatter = logging.Formatter("%(asctime)s %(name)-12s %(levelname)-8s %(message)s")
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)
    root_logger.info("Logger successfully initialised")

    return root_logger


def main() -> None:

    def signal_handler():
        # Handle any cleanup here
        print(
            "SIGINT or CTRL-C detected. Exiting gracefully"
            " and shutting down the dask local cluster"
        )
        client.shutdown()
        sys.exit(0)

    arguments = docopt(__doc__)
    clear_output = arguments["--clear"]
    bucket_name = f's3://{arguments["--input-bucket"]}'
    output_bucket_name = arguments["--output-bucket"]
    outp_dir = arguments["--output-dir"]
    filter_config_file = arguments["--filter-config"]
    output_format = arguments["--format"]
    scheduler = arguments["--scheduler"]
    log_file = arguments["--log-file"]
    nworkers = arguments["--nworkers"] if arguments["--nworkers"] else 8
    log_level = logging.DEBUG if arguments["--verbose"] else logging.INFO
    languages = arguments["--languages"]
    repo_path = arguments["--git-repo"]
    temp_dir = arguments["--temp-dir"]
    prev_manifest_path = (
        arguments["--prev-manifest"] if arguments["--prev-manifest"] else None
    )

    signal.signal(signal.SIGINT, signal_handler)

    if languages:
        languages = languages.split(",")

    init_logging(log_level, log_file)

    # clean output directory if existing
    if outp_dir is not None and os.path.exists(outp_dir):
        if clear_output is not None and clear_output:
            shutil.rmtree(outp_dir)
            os.mkdir(outp_dir)

    with open(filter_config_file, "r", encoding="utf-8") as file:
        config = json.load(file)

    # start the dask local cluster
    if scheduler is None:
        client = Client(n_workers=nworkers, threads_per_worker=1)
    else:
        client = Client(scheduler)

    dask_cluster_msg = f"Dask local cluster: {client}"
    logger.info(dask_cluster_msg)
    print(dask_cluster_msg)

    # the created manifest is not the same based on the output format
    data_stage = "rebuilt" if output_format == "solr" else "passim"

    # initialize manifest
    manifest = DataManifest(
        data_stage=data_stage,
        s3_output_bucket=output_bucket_name,
        s3_input_bucket=bucket_name,
        git_repo=git.Repo(repo_path),
        temp_dir=temp_dir,
        previous_mft_path=prev_manifest_path if prev_manifest_path != "" else None,
    )
    titles = set()

    if arguments["rebuild_articles"]:

        try:
            for n, batch in enumerate(config):
                rebuilt_issues = []
                proc_b_msg = f"Processing batch {n + 1}/{len(config)} [{batch}]"
                logger.info(proc_b_msg)
                print(proc_b_msg)
                newspaper = list(batch.keys())[0]
                start_year, end_year = batch[newspaper]

                for year in range(start_year, end_year):
                    proc_year_msg = f"Processing year {year} \nRetrieving issues..."
                    logger.info(proc_year_msg)
                    print(proc_year_msg)

                    try:
                        input_issues = read_s3_issues(newspaper, year, bucket_name)
                    except FileNotFoundError:
                        fnf_msg = f"{newspaper}-{year} not found in {bucket_name}"
                        logger.info(fnf_msg)
                        print(fnf_msg)
                        continue

                    issue_key, json_files, year_stats = rebuild_issues(
                        issues=input_issues,
                        input_bucket=bucket_name,
                        output_dir=outp_dir,
                        dask_client=client,
                        _format=output_format,
                        filter_language=languages,
                    )
                    rebuilt_issues.append((issue_key, json_files))
                    del input_issues

                    logger.debug("year_stats: %s", year_stats)
                    manifest.add_by_title_year(newspaper, year, year_stats[0])
                    titles.add(newspaper)

                msg = (
                    f"Uploading {len(rebuilt_issues)} rebuilt bz2files "
                    f"to {output_bucket_name}"
                )
                logger.info(msg)
                print(msg)

                b = (
                    db.from_sequence(rebuilt_issues)
                    .starmap(compress, output_dir=outp_dir)
                    .starmap(upload, bucket_name=output_bucket_name)
                    .starmap(cleanup)
                )
                future = b.persist()
                progress(future)
                # clear memory of objects once computations are done
                client.restart()
                rstrt_msg = f"Restarted client after finishing processing batch {n + 1}"
                print(rstrt_msg)
                logger.info(rstrt_msg)

        except Exception as e:
            traceback.print_tb(e.__traceback__)
            print(e)
            client.shutdown()
        finally:
            client.shutdown()

        manifest_note = f"Rebuilt of newspaper articles for {list(titles)}."
        manifest.append_to_notes(manifest_note)
        # finalize and compute the manifest
        manifest.compute(export_to_git_and_s3=False)
        manifest.validate_and_export_manifest(push_to_git=False)

        logger.info("---------- Done ----------")
        print("---------- Done ----------")

    elif arguments["rebuild_pages"]:
        print("\nFunction not yet implemented (sorry!).\n")


if __name__ == "__main__":
    main()



"""Utility functions to export data in Apache UIMA XMI format."""

import os
import json
from typing import Dict, List

from dask import bag as db
from cassis import load_typesystem, Cas


from impresso_commons.utils.s3 import IMPRESSO_STORAGEOPT
from impresso_commons.classes import ContentItem
from impresso_commons.images.olive_boxes import get_iiif_url

IMPRESSO_IIIF_ENDPOINT = "https://dhlabsrv17.epfl.ch/iiif_impresso/"
# IMPRESSO_IIIF_ENDPOINT = 'http://pub.cl.uzh.ch/service/iiif_impresso'


def compute_image_links(
    ci: ContentItem,
    padding: int = 20,
    iiif_endpoint: str = IMPRESSO_IIIF_ENDPOINT,
    iiif_links: Dict[str, str] = None,
    pct: bool = False,
):
    """Short summary.

    :param type ci: Description of parameter `ci`.
    :param type padding: Description of parameter `padding`.
    :return: Description of returned object.
    :rtype: type

    """

    image_links = []
    start_offset = 0

    for line_n, break_offset in enumerate(ci.lines):
        start = start_offset
        end = break_offset
        start_offset = break_offset

        tokens = ci.get_coordinates(start, end)

        if len(tokens) == 0:
            continue

        page_id = tokens[0]["page_id"]

        if "hy1" in tokens[0] and len(tokens) > 1:
            first_token = tokens[1]
        else:
            first_token = tokens[0]

        last_token = tokens[-1]

        if line_n + 1 < len(ci.lines):
            next_offset = ci.lines[line_n + 1]
            next_line_tokens = ci.get_coordinates(start_offset, next_offset)

            if len(next_line_tokens) > 0 and "hy1" in next_line_tokens[0]:
                last_token = next_line_tokens[0]
            else:
                last_token = tokens[-1]

        # compute box coordinates of line
        x1, y1, w1, h1 = first_token["coords"]
        x2, y2, w2, h2 = last_token["coords"]
        x3, y3, w3, h3 = x1, y1 - padding, w2 + (x2 - x1), h1 + padding
        box = " ".join([str(coord) for coord in [x3, y3, w3, h3]])
        if iiif_links is None:
            iiif_link = get_iiif_url(page_id, box, IMPRESSO_IIIF_ENDPOINT, pct)
        else:
            iiif_link = get_iiif_url(
                page_id, box, iiif_manifest_uri=iiif_links[page_id], pct=pct
            )
        image_links.append((iiif_link, start, end))

    return image_links


def get_iiif_links(contentitems: List[ContentItem], canonical_bucket: str):
    """Retrieves from S3 IIIF links for a set of canonical pages where the input content items are found."""

    # derive the IDs of all issues involved
    issue_ids = set(["-".join(ci.id.split("-")[:-1]) for ci in contentitems])

    # reconstruct S3 links to canonical pages
    page_files = [
        os.path.join(
            canonical_bucket,
            issue_id.split("-")[0],
            "pages",
            f"{issue_id.split('-')[0]}-{issue_id.split('-')[1]}",
            f"{issue_id}-pages.jsonl.bz2",
        )
        for issue_id in issue_ids
    ]

    iiif_links = (
        db.read_text(page_files, storage_options=IMPRESSO_STORAGEOPT)
        .map(json.loads)
        .map(lambda x: (x["id"], x["iiif"]))
        .compute()
    )

    return {page_id: iiif_link for page_id, iiif_link in iiif_links}


def rebuilt2xmi(
    ci, output_dir, typesystem_path, iiif_mappings, pct_coordinates=False
) -> str:
    """
    Converts a rebuilt ContentItem into Apache UIMA/XMI format.

    The resulting file will be named after the content item's ID, adding
    the `.xmi` extension.

    :param ci: the content item to be converted
    :type ci: `impresso_commons.classes.ContentItem`
    :param output_dir: the path to the output directory
    :type output_dir: str
    :param typesystem_path: TypeSystem file containing defitions of annotation
        layers.
    :type typesystem_path: str
    """

    with open(typesystem_path, "rb") as f:
        typesystem = load_typesystem(f)

    cas = Cas(typesystem=typesystem)
    cas.sofa_string = ci.fulltext
    cas.sofa_mime = "text/plain"

    sentType = "de.tudarmstadt.ukp.dkpro.core.api.segmentation.type.Sentence"
    imgLinkType = "webanno.custom.ImpressoImages"
    Sentence = typesystem.get_type(sentType)
    ImageLink = typesystem.get_type(imgLinkType)

    # create sentence-level annotations
    start_offset = 0
    for break_offset in ci.lines:
        start = start_offset
        end = break_offset
        start_offset = break_offset
        cas.add_annotation(Sentence(begin=start, end=end))

    iiif_links = compute_image_links(ci, iiif_links=iiif_mappings, pct=pct_coordinates)

    # inject the IIIF links into
    for iiif_link, start, end in iiif_links:
        cas.add_annotation(ImageLink(begin=start, end=end, link=iiif_link))

    outfile_path = os.path.join(output_dir, f"{ci.id}.xmi")
    cas.to_xmi(outfile_path, pretty_print=True)
    return outfile_path



#!/usr/bin/env python
# coding: utf-8

import json
import os
import logging
import pathlib
from contextlib import ExitStack
from typing import Any, Optional
import jsonschema
import importlib_resources

logger = logging.getLogger(__name__)


def parse_json(filename):
    if os.path.isfile(filename):
        with open(filename, "r") as f:
            return json.load(f)
    else:
        logger.info(f"File {filename} does not exist.")


def chunk(list, chunksize):
    """Yield successive n-sized chunks from list."""
    for i in range(0, len(list), chunksize):
        yield list[i : i + chunksize]


def get_pkg_resource(
    file_manager: ExitStack, path: str, package: str = "impresso_commons"
) -> pathlib.PosixPath:
    """Return the resource at `path` in `package`, using a context manager.

    Note:
        The context manager `file_manager` needs to be instantiated prior to
        calling this function and should be closed once the package resource
        is no longer of use.

    Args:
        file_manager (contextlib.ExitStack): Context manager.
        path (str): Path to the desired resource in given package.
        package (str, optional): Package name. Defaults to "impresso_commons".

    Returns:
        pathlib.PosixPath: Path to desired managed resource.
    """
    ref = importlib_resources.files(package) / path
    return file_manager.enter_context(importlib_resources.as_file(ref))


def init_logger(
    level: int = logging.INFO, file: Optional[str] = None
) -> logging.RootLogger:
    """Initialises the root logger.

    Args:
        level (int, optional): desired level of logging. Defaults to logging.INFO.
        file (str | None, optional): _description_. Defaults to None.

    Returns:
        logging.RootLogger: the initialised logger
    """
    # Initialise the logger
    root_logger = logging.getLogger("")
    root_logger.setLevel(level)

    if file is not None:
        handler = logging.FileHandler(filename=file, mode="w")
    else:
        handler = logging.StreamHandler()

    formatter = logging.Formatter("%(asctime)s %(name)-12s %(levelname)-8s %(message)s")
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)
    root_logger.info("Logger successfully initialised")

    return root_logger


def validate_against_schema(
    json_to_validate: dict[str, Any],
    path_to_schema: str = "schemas/json/versioning/manifest.schema.json",
) -> None:
    """Validate a dict corresponding to a JSON against a provided JSON schema.

    Args:
        json (dict[str, Any]): JSON data to validate against a schema.
        path_to_schema (str, optional): Path to the JSON schema to validate against.
            Defaults to "impresso-schemas/json/versioning/manifest.schema.json".

    Raises:
        e: The provided JSON could not be validated against the provided schema.
    """
    file_manager = ExitStack()
    schema_path = get_pkg_resource(file_manager, path_to_schema)
    with open(os.path.join(schema_path), "r", encoding="utf-8") as f:
        json_schema = json.load(f)

    try:
        jsonschema.validate(json_to_validate, json_schema)
    except Exception as e:
        logger.error(
            "The provided JSON could not be validated against its schema: %s.",
            json_to_validate,
        )
        raise e


def bytes_to(bytes_nb: int, to_unit: str, bsize: int = 1024) -> float:
    """Convert bytes to the specified unit.

    Supported target units:
    - 'k' (kilobytes), 'm' (megabytes),
    - 'g' (gigabytes), 't' (terabytes),
    - 'p' (petabytes), 'e' (exabytes).

    Args:
        bytes_nb (int): The number of bytes to be converted.
        to_unit (str): The target unit for conversion.
        bsize (int, optional): The base size used for conversion (default is 1024).

    Returns:
        float: The converted value in the specified unit.

    Raises:
        KeyError: If the specified target unit is not supported.
    """
    units = {"k": 1, "m": 2, "g": 3, "t": 4, "p": 5, "e": 6}
    return float(bytes_nb) / (bsize ** units[to_unit])



"""This module contains the definition of a data statistics class.

A DataStatstics object should be instantiated during each processing step of 
the data preprocessing and augmentation of the Impresso project, and used to 
progressively count the number of elements modified or added by the processing.
"""

import logging
from abc import ABC, abstractmethod
from typing import Any, Union, Self, Optional

# from impresso_commons.versioning.data_manifest import DataStage
from impresso_commons.versioning.helpers import (
    DataStage,
    validate_stage,
    validate_granularity,
)

logger = logging.getLogger(__name__)

POSSIBLE_ACTIONS = ["addition", "modification"]
POSSIBLE_GRANULARITIES = ["corpus", "title", "year"]


class DataStatistics(ABC):
    """Count statistics computed on a specific portion and granularity of the data.

    Args:
        data_stage (DataStage | str): The stage of data the stats are computed on.
        granularity (str): The granularity of the statistics with respect to the data.
        element (str, optional): The specific element associated with the statistics.
            Defaults to "" (empty string).
        counts (dict[str, int | dict[str, int]] | None, optional): Initial counts for
            statistics. Defaults to None.

    Attributes:
        stage (DataStage): The stage of data the stats are computed on.
        granularity (str): The granularity of the statistics with respect to the data.
        element (str): The specific element associated with the statistics.
        count_keys (list[str]): The count keys for these statistics.
        counts (dict[str, int | dict[str, int]]): The count statistics computed on the
            specific data, can include frequency dicts.
    """

    def __init__(
        self,
        data_stage: Union[DataStage, str],
        granularity: str,
        element: Optional[str] = None,
        counts: Union[dict[str, Union[int, dict[str, int]]], None] = None,
    ) -> None:

        self.stage = validate_stage(data_stage)
        self.granularity = validate_granularity(granularity)
        self.element = element
        self.count_keys = self._define_count_keys()

        if counts is not None and self._validate_count_keys(counts):
            self.counts = counts
        else:
            logger.debug("Initializing counts to 0 for %s.", self.element)
            self.counts = self.init_counts()

    @abstractmethod
    def _define_count_keys(self) -> list[str]:
        """Define the count keys for these specific statistics."""

    @abstractmethod
    def _validate_count_keys(
        self, new_counts: dict[str, Union[int, dict[str, int]]]
    ) -> bool:
        """Validate the keys of new counts provided against defined count keys."""

    def init_counts(self) -> dict[str, Union[int, dict[str, int]]]:
        """Initialize a dict with all the keys associated to this object.

        Returns:
            dict[str, int | dict[str, int]]: A dict with all defined keys, and values
                initialized to 0 (or to empty frequency dicts).
        """
        return {k: 0 if "fd" not in k else {} for k in self.count_keys}

    def add_counts(
        self, new_counts: dict[str, Union[int, dict[str, int]]], replace: bool = False
    ) -> bool:
        """Add new counts to the existing counts if the new keys are validated.

        Args:
            new_counts (dict[str, int | dict[str, int]]): New counts to be added.

        Returns:
            bool: True if the counts were valid and could be added, False otherwise.
        """
        if self._validate_count_keys(new_counts):
            if replace:
                logger.debug(
                    "Replacing the counts by %s for %s. This will erase previous counts.",
                    new_counts,
                    self.element,
                )
                self.counts = new_counts
            else:
                for k, v in new_counts.items():
                    if k.endswith("_fd"):
                        # some fields can be frequency dicts
                        for v_k, v_f in v.items():
                            self.counts[k][v_k] = (
                                v_f
                                if v_k not in self.counts[k]
                                else self.counts[k][v_k] + v_f
                            )
                    else:
                        self.counts[k] += v
            return True

        return False

    def pretty_print(
        self, modif_date: Optional[str] = None, include_counts: bool = False
    ) -> dict[str, Any]:
        """Generate a dict representation of these statistics to add to a json.

        These stats are agnostic to the type of statistics they represent so the values
        of `self.counts` are excluded by default, to be included in child classes.
        The modification date can also be included (when granularity='year')

        Args:
            modif_date (Optional[str], optional): Last modification date of the
                corresponding elements. Defaults to None.
            include_counts (bool, optional): Whether to include the current counts with
                key "stats". Defaults to False.

        Returns:
            dict[str, Any]: A dict with the general information about these statistics.
        """
        stats_dict = {
            "stage": self.stage.value,
            "granularity": self.granularity,
        }

        # no element for the overall stats
        if self.granularity != "corpus":
            if self.element is not None:
                stats_dict["element"] = self.element
            else:
                logger.warning("Missing the element when pretty-printing!")

        # If a modification date is provided, add it.
        if modif_date is not None:
            stats_dict["last_modification_date"] = modif_date
            if self.granularity != "year":
                # if the granularity is not year, log a warning (unexpected behavior)
                logger.warning(
                    "'last_modification_date' field was added although granularity is %s",
                    self.granularity,
                )

        if include_counts:
            stats_dict["stats"] = {
                k: (
                    v
                    if "fd" not in k
                    else {v_k: v_f for v_k, v_f in v.items() if v_f > 0}
                )
                for k, v in self.counts.items()
                if "_fd" in k or v > 0
            }

        return stats_dict

    @abstractmethod
    def same_counts(self, other_stats: Union[dict[str, Any], Self]) -> bool:
        """Given another dict of stats, check whether the values are the same."""


class NewspaperStatistics(DataStatistics):
    """Count statistics computed on a specific portion and granularity of the data.

    Args:
        data_stage (DataStage | str): The stage of data the stats are computed on.
        granularity (str): The granularity of the statistics with respect to the data.
        element (str, optional): The specific element associated with the statistics.
            Defaults to "" (empty string).
        counts (dict[str, int] | None, optional): Initial counts for statistics.
            Defaults to None.

    Attributes:
        stage (DataStage): The stage of data the stats are computed on.
        granularity (str): The granularity of the statistics with respect to the data.
        element (str): The specific element associated with the statistics.
        count_keys (list[str]): The count keys for these statistics.
        counts (dict[str, int]): The count statistics computed on the specific data.
    """

    # All possible count keys for newspaper data.
    possible_count_keys = [
        "titles",
        "issues",
        "pages",
        "content_items_out",
        "ft_tokens",
        "images",
        "content_items_in",
        "ne_mentions",
        "ne_entities",
        "embeddings_el",
        "topics",
        "lang_fd",  # '_fd' suffix signifies a frenquency dict
        "text_reuse_clusters",
        "text_reuse_passages",
    ]

    def _define_count_keys(self) -> list[str]:
        """Define the count keys to use for these specific statistics.

        TODO correct/update the count_keys

        Returns:
            list[str]: The count keys for this specific stage and granularity.
        """
        start_index = int(self.granularity != "corpus")
        # all counts should have 'content_items_out'
        count_keys = [self.possible_count_keys[3]]
        # add 'issues' and 'titles' (only if corpus granularity)
        count_keys.extend(self.possible_count_keys[start_index:2])

        match self.stage:
            case DataStage.CANONICAL:
                # add 'pages' and 'images'
                count_keys.append(self.possible_count_keys[2])
                count_keys.append(self.possible_count_keys[5])
                # keys: 'content_items_out', 'titles', 'issues', 'pages', 'images'
            case DataStage.REBUILT:
                # add 'ft_tokens'
                count_keys.append(self.possible_count_keys[4])
                # keys: 'content_items_out', 'titles', 'issues', 'ft_tokens'
            case DataStage.EMBEDDINGS:
                # add 'embeddings'
                count_keys.append(self.stage.value)
                # TODO update
                # keys: 'content_items_out', 'titles', 'issues', 'embeddings'
            case DataStage.ENTITIES:
                # add 'ne_entities', 'ne_mentions'
                count_keys.extend(self.possible_count_keys[7:9])
                # keys: 'content_items_out', 'titles', 'issues', 'ne_entities', 'ne_mentions'
            case DataStage.PASSIM:
                # add 'ft_tokens'
                count_keys.append(self.possible_count_keys[4])
                # keys: 'content_items_out', 'titles', 'issues', 'ft_tokens'
            case DataStage.LANGIDENT:
                # add 'images', 'lang_fd'
                count_keys.append(self.possible_count_keys[5])
                count_keys.append(self.possible_count_keys[11])
                # keys: 'content_items_out', 'titles', 'issues', 'pages', 'lang_fd'
            case DataStage.TEXT_REUSE:
                # add 'text_reuse_clusters' and 'text_reuse_passages'
                count_keys.append(self.possible_count_keys[12])
                count_keys.append(self.possible_count_keys[13])
                # keys: 'content_items_out', 'titles', 'issues', 'text_reuse_clusters', 'text_reuse_passages'
            case DataStage.TOPICS:
                # add 'topics'
                # TODO update
                count_keys.append(self.possible_count_keys[10])
            case DataStage.MYSQL_CIS:
                # add 'pages'
                count_keys.append(self.possible_count_keys[2])
                # keys: 'content_items_out', 'titles', 'issues', 'pages'

        # For case DataStage.SOLR_TEXT, all keys are already added.
        #   keys: 'content_items_out', 'titles', 'issues'
        return count_keys

    def _validate_count_keys(
        self, new_counts: dict[str, Union[int, dict[str, int]]]
    ) -> bool:
        """Validate the keys of new counts provided against defined count keys.

        Valid new counts shouldn't have keys absent from the defined `attr:count_keys`
        or non-integer values.

        Args:
            new_counts (dict[str, int | dict[str, int]]): New counts to validate

        Returns:
            bool: True if `new_counts` are valid, False otherwise.
        """
        if not all(k in self.count_keys for k in new_counts.keys()):
            warn_msg = (
                f"Provided value `counts`: {new_counts} has keys not present in "
                f"`count_keys`: {self.count_keys}. The counts provided won't be used."
            )
            logger.error(warn_msg)
            return False

        if not all(
            v >= 0 if "fd" not in k else all(fv > 0 for fv in v.values())
            for k, v in new_counts.items()
        ):
            logger.error(
                "Provided count values are not all integers and will not be used."
            )
            return False

        # the provided counts were conforming
        return True

    def pretty_print(
        self, modif_date: Optional[str] = None, include_counts: bool = True
    ) -> dict[str, Any]:
        """Generate a dict representation of these statistics to add to a json.

        Args:
            modif_date (Optional[str], optional): Last modification date of the
                corresponding elements. Defaults to None.
            include_counts (bool, optional): Whether to include the current newspaper
                counts with key "nps_stats". Defaults to True.

        Returns:
            dict[str, Any]: A dict representation of these statistics.
        """
        stats_dict = super().pretty_print(modif_date=modif_date)
        # add the newspaper stats
        if include_counts:
            stats_dict["nps_stats"] = {
                k: (
                    v
                    if "_fd" not in k
                    else {v_k: v_f for v_k, v_f in v.items() if v_f > 0}
                )
                for k, v in self.counts.items()
                if "_fd" in k or v > 0
            }

        return stats_dict

    def same_counts(self, other_stats: Union[dict[str, Any], Self]) -> bool:
        """Given another dict of stats, check whether the values are the same.

        Args:
            other_stats (Union[dict[str, Any], Self]): Dict with pretty-printed
                newspaper statistics or other NewspaperStatistics object.

        Returns:
            bool: True if the values for the various fields of `nps_stats` where the
                same, False otherwise.
        """
        if isinstance(other_stats, NewspaperStatistics):
            other_stats = other_stats.pretty_print()

        self_stats = self.pretty_print()
        return self_stats["nps_stats"] == other_stats["nps_stats"]
